# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module defines point regions in both pixel and sky coordinates.
"""

import numpy as np

from regions.core.attributes import (RegionMetaDescr, RegionVisualDescr,
                                     ScalarPixCoord, ScalarSkyCoord)
from regions.core.bounding_box import RegionBoundingBox
from regions.core.core import PixelRegion, SkyRegion
from regions.core.metadata import RegionMeta, RegionVisual
from regions.core.pixcoord import PixCoord

__all__ = ['PointPixelRegion', 'PointSkyRegion']


class PointPixelRegion(PixelRegion):
    """
    A point position in pixel coordinates.

    Parameters
    ----------
    center : `~regions.PixCoord`
        The position of the point.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.

    Examples
    --------
    .. plot::
        :include-source:

        from regions import PixCoord, PointPixelRegion, RegionVisual
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(1, 1)

        regs = []
        regs.append(PointPixelRegion(PixCoord(2, 2),
                    visual=RegionVisual(marker='D')))
        regs.append(PointPixelRegion(PixCoord(2, 3),
                    visual=RegionVisual(marker='+')))
        regs.append(PointPixelRegion(PixCoord(3, 3),
                    visual=RegionVisual(marker='^')))
        regs.append(PointPixelRegion(PixCoord(3, 2),
                    visual=RegionVisual(marker='*')))
        regs.append(PointPixelRegion(PixCoord(2, 4),
                    visual=RegionVisual(marker='x')))
        regs.append(PointPixelRegion(PixCoord(4, 2)))
        for reg in regs:
            reg.plot(ax=ax)

        ax.set_xlim(0, 6)
        ax.set_ylim(0, 6)
        ax.set_aspect('equal')
    """

    _params = ('center',)
    _mpl_artist = 'Line2D'
    center = ScalarPixCoord('The point pixel position as a |PixCoord|.')
    meta = RegionMetaDescr('The meta attributes as a |RegionMeta|')
    visual = RegionVisualDescr('The visual attributes as a |RegionVisual|.')

    def __init__(self, center, meta=None, visual=None):
        self.center = center
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

    @property
    def area(self):
        return 0.0

    def contains(self, pixcoord):
        in_reg = (False if pixcoord.isscalar
                  else np.zeros(pixcoord.x.shape, dtype=bool))

        if self.meta.get('include', True):
            # in_reg = False, always.  Points do not include anything.
            return in_reg
        else:
            return np.logical_not(in_reg)

    def to_sky(self, wcs):
        center = wcs.pixel_to_world(self.center.x, self.center.y)
        return PointSkyRegion(center, meta=self.meta.copy(),
                              visual=self.visual.copy())

    @property
    def bounding_box(self):
        return RegionBoundingBox.from_float(self.center.x, self.center.x,
                                            self.center.y, self.center.y)

    def to_mask(self, mode='center', subpixels=5):
        # TODO: needs to be implemented
        raise NotImplementedError

    def as_artist(self, origin=(0, 0), **kwargs):
        """
        Return a matplotlib Line2D object for this region
        (`matplotlib.lines.Line2D`).

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` pixel position of the origin of the displayed
            image.

        **kwargs : `dict`
            Any keyword arguments accepted by
            `~matplotlib.lines.Line2D`. These keywords will override any
            visual meta attributes of this region.

        Returns
        -------
        artist : `~matplotlib.lines.Line2D`
            A matplotlib Line2D object.
        """
        from matplotlib.lines import Line2D

        mpl_kwargs = self.visual.define_mpl_kwargs(self._mpl_artist)
        mpl_kwargs.update(kwargs)

        return Line2D([self.center.x - origin[0]],
                      [self.center.y - origin[1]], **mpl_kwargs)

    def rotate(self, center, angle):
        """
        Rotate the region.

        Positive ``angle`` corresponds to counter-clockwise rotation.

        Parameters
        ----------
        center : `~regions.PixCoord`
            The rotation center point.
        angle : `~astropy.coordinates.Angle`
            The rotation angle.

        Returns
        -------
        region : `PointPixelRegion`
            The rotated region (which is an independent copy).
        """
        center = self.center.rotate(center, angle)
        return self.copy(center=center)


class PointSkyRegion(SkyRegion):
    """
    A pixel region in sky coordinates.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The position of the point.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.
    """

    _params = ('center',)
    center = ScalarSkyCoord('The point position as a |SkyCoord|.')
    meta = RegionMetaDescr('The meta attributes as a |RegionMeta|')
    visual = RegionVisualDescr('The visual attributes as a |RegionVisual|.')

    def __init__(self, center, meta=None, visual=None):
        self.center = center
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

    def contains(self, skycoord, wcs):  # pylint: disable=unused-argument
        # points never include anything
        return not self.meta.get('include', True)

    def to_pixel(self, wcs):
        center_x, center_y = wcs.world_to_pixel(self.center)
        center = PixCoord(center_x, center_y)
        return PointPixelRegion(center, meta=self.meta.copy(),
                                visual=self.visual.copy())



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.data import get_pkg_data_filename
from astropy.wcs import WCS
from numpy.testing import assert_allclose

from regions._utils.optional_deps import HAS_MATPLOTLIB
from regions.core import PixCoord, RegionMeta, RegionVisual
from regions.shapes.circle import CirclePixelRegion, CircleSkyRegion
from regions.shapes.tests.test_common import (BaseTestPixelRegion,
                                              BaseTestSkyRegion)
from regions.tests.helpers import make_simple_wcs


@pytest.fixture(scope='session', name='wcs')
def wcs_fixture():
    filename = get_pkg_data_filename('data/example_header.fits')
    header = fits.getheader(filename)
    return WCS(header)


class TestCirclePixelRegion(BaseTestPixelRegion):
    meta = RegionMeta({'text': 'test'})
    visual = RegionVisual({'color': 'blue'})
    reg = CirclePixelRegion(PixCoord(3, 4), radius=2, meta=meta, visual=visual)
    sample_box = [0, 6, 1, 7]
    inside = [(3, 4)]
    outside = [(3, 0)]
    expected_area = 4 * np.pi
    expected_repr = '<CirclePixelRegion(center=PixCoord(x=3, y=4), radius=2)>'
    expected_str = ('Region: CirclePixelRegion\ncenter: PixCoord(x=3, y=4)'
                    '\nradius: 2')

    def test_copy(self):
        reg = self.reg.copy()
        assert reg.center.xy == (3, 4)
        assert reg.radius == 2
        assert reg.meta == self.meta
        assert reg.visual == self.visual

    def test_pix_sky_roundtrip(self):
        wcs = make_simple_wcs(SkyCoord(2 * u.deg, 3 * u.deg), 0.1 * u.deg, 20)
        reg_new = self.reg.to_sky(wcs).to_pixel(wcs)
        assert_allclose(reg_new.center.x, self.reg.center.x)
        assert_allclose(reg_new.center.y, self.reg.center.y)
        assert_allclose(reg_new.radius, self.reg.radius)
        assert reg_new.meta == self.reg.meta
        assert reg_new.visual == self.reg.visual

        # test that converted region meta and visual are copies and not views
        reg_new.meta['text'] = 'new'
        reg_new.visual['color'] = 'green'
        assert reg_new.meta['text'] != self.reg.meta['text']
        assert reg_new.visual['color'] != self.reg.visual['color']

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_as_artist(self):
        patch = self.reg.as_artist()
        assert_allclose(patch.center, (3, 4))
        assert_allclose(patch.radius, 2)

    def test_rotate(self):
        reg = self.reg.rotate(PixCoord(2, 3), 90 * u.deg)
        assert_allclose(reg.center.xy, (1, 4))

    def test_eq(self):
        reg = self.reg.copy()
        assert reg == self.reg
        reg.radius = 3
        assert reg != self.reg

    def test_zero_size(self):
        with pytest.raises(ValueError):
            CirclePixelRegion(PixCoord(50, 50), radius=0)


class TestCircleSkyRegion(BaseTestSkyRegion):
    meta = RegionMeta({'text': 'test'})
    visual = RegionVisual({'color': 'blue'})
    reg = CircleSkyRegion(SkyCoord(3 * u.deg, 4 * u.deg), 2 * u.arcsec,
                          meta=meta, visual=visual)

    expected_repr = ('<CircleSkyRegion(center=<SkyCoord (ICRS): (ra, dec) in '
                     'deg\n    (3., 4.)>, radius=2.0 arcsec)>')
    expected_str = ('Region: CircleSkyRegion\ncenter: <SkyCoord (ICRS): '
                    '(ra, dec) in deg\n    (3., 4.)>\nradius: 2.0 '
                    'arcsec')

    def test_copy(self):
        reg = self.reg.copy()
        assert_allclose(reg.center.ra.deg, 3)
        assert_allclose(reg.radius.to_value('arcsec'), 2)
        assert reg.meta == self.meta
        assert reg.visual == self.visual

    def test_transformation(self, wcs):
        skycoord = SkyCoord(3 * u.deg, 4 * u.deg, frame='galactic')
        skycircle = CircleSkyRegion(skycoord, 2 * u.arcsec)

        pixcircle = skycircle.to_pixel(wcs)

        assert_allclose(pixcircle.center.x, -50.5)
        assert_allclose(pixcircle.center.y, 299.5)
        assert_allclose(pixcircle.radius, 0.027777777777828305)

        skycircle2 = pixcircle.to_sky(wcs)

        assert_quantity_allclose(skycircle.center.data.lon,
                                 skycircle2.center.data.lon)
        assert_quantity_allclose(skycircle.center.data.lat,
                                 skycircle2.center.data.lat)
        assert_quantity_allclose(skycircle2.radius, skycircle.radius)

    def test_dimension_center(self):
        center = SkyCoord([1, 2] * u.deg, [3, 4] * u.deg)
        radius = 2 * u.arcsec
        with pytest.raises(ValueError) as excinfo:
            CircleSkyRegion(center, radius)
        estr = "'center' must be a scalar SkyCoord"
        assert estr in str(excinfo.value)

    def test_contains(self, wcs):
        position = SkyCoord([1, 3] * u.deg, [2, 4] * u.deg)
        # 1,2 is outside, 3,4 is the center and is inside
        assert all(self.reg.contains(position, wcs)
                   == np.array([False, True], dtype='bool'))

    def test_eq(self):
        reg = self.reg.copy()
        assert reg == self.reg
        reg.radius = 3 * u.arcsec
        assert reg != self.reg

    def test_zero_size(self):
        with pytest.raises(ValueError):
            CircleSkyRegion(SkyCoord(3 * u.deg, 4 * u.deg), 0. * u.arcsec)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from pathlib import Path

import click
import pyregion
from astropy import log

from regions import read_ds9

log.setLevel('DEBUG')

TEST_FILE_DIR = Path('../regions/io/ds9/tests/data')


@click.group()
def cli():
    """
    Region parser debugging tool.
    """
    pass


@cli.command('list-files')
def list_files():
    print('Available files')
    for ffile in TEST_FILE_DIR.glob('*.reg'):
        print(ffile.parts[-1])


@cli.command('parse')
@click.option('--interactive', is_flag=True, default=False)
@click.option('--parser', default='regions')
@click.option('--errors', default='strict')
@click.argument('filename')
def parse(filename, interactive, parser, errors):
    readname = TEST_FILE_DIR / filename
    print(f'Reading {readname}')
    print(f'Using parser {parser}')
    if parser == 'regions':
        regions = read_ds9(str(readname), errors=errors)
    elif parser == 'pyregion':
        regions = pyregion.open(str(readname))
    print(regions)
    if interactive:
        import IPython
        IPython.embed()


if __name__ == '__main__':
    cli()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module defines line regions in both pixel and sky coordinates.
"""

import numpy as np

from regions.core.attributes import (RegionMetaDescr, RegionVisualDescr,
                                     ScalarPixCoord, ScalarSkyCoord)
from regions.core.bounding_box import RegionBoundingBox
from regions.core.core import PixelRegion, SkyRegion
from regions.core.metadata import RegionMeta, RegionVisual
from regions.core.pixcoord import PixCoord

__all__ = ['LinePixelRegion', 'LineSkyRegion']


class LinePixelRegion(PixelRegion):
    """
    A line in pixel coordinates.

    Parameters
    ----------
    start : `~regions.PixCoord`
        The start position.
    end : `~regions.PixCoord`
        The end position.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.

    Examples
    --------
    .. plot::
        :include-source:

        from regions import PixCoord, LinePixelRegion
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(1, 1)

        start = PixCoord(x=15, y=10)
        end = PixCoord(x=20, y=25)
        reg = LinePixelRegion(start=start, end=end)
        patch = reg.plot(ax=ax, edgecolor='red', lw=2, label='Line')

        ax.legend(handles=(patch,), loc='upper center')
        ax.set_xlim(0, 30)
        ax.set_ylim(0, 30)
        ax.set_aspect('equal')
    """

    _params = ('start', 'end')
    _mpl_artist = 'Patch'
    start = ScalarPixCoord('The start pixel position as a |PixCoord|.')
    end = ScalarPixCoord('The end pixel position as a |PixCoord|.')
    meta = RegionMetaDescr('The meta attributes as a |RegionMeta|')
    visual = RegionVisualDescr('The visual attributes as a |RegionVisual|.')

    def __init__(self, start, end, meta=None, visual=None):
        self.start = start
        self.end = end
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

    @property
    def area(self):
        return 0

    def contains(self, pixcoord):
        in_reg = (False if pixcoord.isscalar
                  else np.zeros(pixcoord.x.shape, dtype=bool))

        if self.meta.get('include', True):
            return in_reg
        else:
            return np.logical_not(in_reg)

    def to_sky(self, wcs):
        start = wcs.pixel_to_world(self.start.x, self.start.y)
        end = wcs.pixel_to_world(self.end.x, self.end.y)
        return LineSkyRegion(start, end, meta=self.meta.copy(),
                             visual=self.visual.copy())

    @property
    def bounding_box(self):
        xmin = min(self.start.x, self.end.x)
        xmax = max(self.start.x, self.end.x)
        ymin = min(self.start.y, self.end.y)
        ymax = max(self.start.y, self.end.y)

        return RegionBoundingBox.from_float(xmin, xmax, ymin, ymax)

    def to_mask(self, mode='center', subpixels=5):
        # TODO: needs to be implemented
        raise NotImplementedError

    def as_artist(self, origin=(0, 0), **kwargs):
        """
        Return a matplotlib patch object for this region
        (`matplotlib.patches.Arrow`).

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` pixel position of the origin of the displayed
            image.

        **kwargs : dict
            Any keyword arguments accepted by
            `~matplotlib.patches.Arrow`. These keywords will override
            any visual meta attributes of this region.

        Returns
        -------
        artist : `~matplotlib.patches.Arrow`
            A matplotlib line patch.
        """
        # Note: Long term we want to support DS9 lines with arrow heads.
        # We may want to use Line2D instead of arrow for lines because the
        # width of the arrow is non-scalable in patches.
        from matplotlib.patches import Arrow

        x = self.start.x - origin[0]
        y = self.start.y - origin[1]
        dx = self.end.x - self.start.x
        dy = self.end.y - self.start.y
        kwargs.setdefault('width', 0.1)

        mpl_kwargs = self.visual.define_mpl_kwargs(self._mpl_artist)
        mpl_kwargs.update(kwargs)

        return Arrow(x, y, dx, dy, **mpl_kwargs)

    def rotate(self, center, angle):
        """
        Rotate the region.

        Positive ``angle`` corresponds to counter-clockwise rotation.

        Parameters
        ----------
        center : `~regions.PixCoord`
            The rotation center point.
        angle : `~astropy.coordinates.Angle`
            The rotation angle.

        Returns
        -------
        region : `LinePixelRegion`
            The rotated region (which is an independent copy).
        """
        start = self.start.rotate(center, angle)
        end = self.end.rotate(center, angle)
        return self.copy(start=start, end=end)


class LineSkyRegion(SkyRegion):
    """
    A line in sky coordinates.

    Parameters
    ----------
    start : `~astropy.coordinates.SkyCoord`
        The start position.
    end : `~astropy.coordinates.SkyCoord`
        The end position.
    meta : `~regions.RegionMeta` or `dict`, optional
        A dictionary that stores the meta attributes of the region.
    visual : `~regions.RegionVisual` or `dict`, optional
        A dictionary that stores the visual meta attributes of the
        region.
    """

    _params = ('start', 'end')
    start = ScalarSkyCoord('The start position as a |SkyCoord|.')
    end = ScalarSkyCoord('The end position as a |SkyCoord|.')
    meta = RegionMetaDescr('The meta attributes as a |RegionMeta|')
    visual = RegionVisualDescr('The visual attributes as a |RegionVisual|.')

    def __init__(self, start, end, meta=None, visual=None):
        self.start = start
        self.end = end
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

    def contains(self, skycoord, wcs):  # pylint: disable=unused-argument
        # lines never contain anything
        return not self.meta.get('include', True)

    def to_pixel(self, wcs):
        start_x, start_y = wcs.world_to_pixel(self.start)
        start = PixCoord(start_x, start_y)
        end_x, end_y = wcs.world_to_pixel(self.end)
        end = PixCoord(end_x, end_y)
        return LinePixelRegion(start, end, meta=self.meta.copy(),
                               visual=self.visual.copy())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the metadata module.
"""
import pytest

from regions.core.metadata import RegionMeta, RegionVisual


def test_region_meta():
    meta_dict = {'text': 'hello world', 'tag': ['Tag1', 'Tag2']}
    meta = RegionMeta(meta_dict)

    for key, val in meta_dict.items():
        assert val == meta[key]

    meta2 = meta.copy()
    assert isinstance(meta2, RegionMeta)

    text = 'new'
    meta['text'] = text
    assert meta2['text'] != text

    with pytest.raises(KeyError):
        RegionMeta({'invalid': 1})
    with pytest.raises(KeyError):
        meta.update({'invalid': 1})
    with pytest.raises(KeyError):
        meta.update(invalid=1)
    with pytest.raises(KeyError):
        meta.setdefault('invalid', 1)


def test_region_visual():
    meta_dict = {'color': 'blue', 'fontsize': 12}
    meta = RegionVisual(meta_dict)

    for key, val in meta_dict.items():
        assert val == meta[key]

    meta2 = meta.copy()
    assert isinstance(meta2, RegionVisual)

    color = 'green'
    meta['color'] = color
    assert meta2['color'] != color

    with pytest.raises(KeyError):
        RegionVisual({'invalid': 1})
    with pytest.raises(KeyError):
        meta.update({'invalid': 1})
    with pytest.raises(KeyError):
        meta.update(invalid=1)
    with pytest.raises(KeyError):
        meta.setdefault('invalid', 1)


def test_region_visual_mpl_kwargs():
    meta_dict = {'color': 'blue'}
    meta = RegionVisual(meta_dict)

    kwargs = meta.define_mpl_kwargs('Patch')
    expected = {'edgecolor': 'blue', 'fill': False}
    assert kwargs == expected

    kwargs = meta.define_mpl_kwargs('Line2D')
    expected = {'markeredgecolor': 'blue', 'fillstyle': 'none',
                'marker': 'o'}
    assert kwargs == expected

    kwargs = meta.define_mpl_kwargs('Text')
    expected = {'color': 'blue'}
    assert kwargs == expected

    with pytest.raises(ValueError):
        meta.define_mpl_kwargs('invalid')



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the bounding_box module.
"""

import pytest
from numpy.testing import assert_allclose

from regions._utils.optional_deps import HAS_MATPLOTLIB
from regions.core.bounding_box import RegionBoundingBox
from regions.core.pixcoord import PixCoord
from regions.shapes import RectanglePixelRegion


def test_bounding_box_init():
    bbox = RegionBoundingBox(1, 10, 2, 20)
    assert bbox.ixmin == 1
    assert bbox.ixmax == 10
    assert bbox.iymin == 2
    assert bbox.iymax == 20


def test_bounding_box_init_minmax():
    with pytest.raises(ValueError):
        RegionBoundingBox(100, 1, 1, 100)
    with pytest.raises(ValueError):
        RegionBoundingBox(1, 100, 100, 1)


def test_bounding_box_inputs():
    with pytest.raises(TypeError):
        RegionBoundingBox([1], [10], [2], [9])
    with pytest.raises(TypeError):
        RegionBoundingBox([1, 2], 10, 2, 9)
    with pytest.raises(TypeError):
        RegionBoundingBox(1.0, 10.0, 2.0, 9.0)
    with pytest.raises(TypeError):
        RegionBoundingBox(1.3, 10, 2, 9)
    with pytest.raises(TypeError):
        RegionBoundingBox(1, 10.3, 2, 9)
    with pytest.raises(TypeError):
        RegionBoundingBox(1, 10, 2.3, 9)
    with pytest.raises(TypeError):
        RegionBoundingBox(1, 10, 2, 9.3)


def test_bounding_box_from_float():
    # This is the example from the method docstring
    bbox = RegionBoundingBox.from_float(xmin=1.0, xmax=10.0, ymin=2.0,
                                        ymax=20.0)
    assert bbox == RegionBoundingBox(ixmin=1, ixmax=11, iymin=2, iymax=21)

    bbox = RegionBoundingBox.from_float(xmin=1.4, xmax=10.4, ymin=1.6,
                                        ymax=10.6)
    assert bbox == RegionBoundingBox(ixmin=1, ixmax=11, iymin=2, iymax=12)


def test_bounding_box_eq():
    bbox = RegionBoundingBox(1, 10, 2, 20)
    assert bbox == RegionBoundingBox(1, 10, 2, 20)
    assert bbox != RegionBoundingBox(9, 10, 2, 20)
    assert bbox != RegionBoundingBox(1, 99, 2, 20)
    assert bbox != RegionBoundingBox(1, 10, 9, 20)
    assert bbox != RegionBoundingBox(1, 10, 2, 99)

    with pytest.raises(TypeError):
        assert bbox == (1, 10, 2, 20)


def test_bounding_box_repr():
    bbox = RegionBoundingBox(1, 10, 2, 20)
    assert repr(bbox) == ('RegionBoundingBox(ixmin=1, ixmax=10, iymin=2, '
                          'iymax=20)')


def test_bounding_box_shape():
    bbox = RegionBoundingBox(1, 10, 2, 20)
    assert bbox.shape == (18, 9)


def test_bounding_box_center():
    bbox = RegionBoundingBox(1, 10, 2, 20)
    assert bbox.center == (10.5, 5)


def test_bounding_box_get_overlap_slices():
    bbox = RegionBoundingBox(1, 10, 2, 20)
    slc = ((slice(2, 20, None), slice(1, 10, None)),
           (slice(0, 18, None), slice(0, 9, None)))
    assert bbox.get_overlap_slices((50, 50)) == slc

    bbox = RegionBoundingBox(-10, -1, 2, 20)
    assert bbox.get_overlap_slices((50, 50)) == (None, None)

    bbox = RegionBoundingBox(-10, 10, -10, 20)
    slc = ((slice(0, 20, None), slice(0, 10, None)),
           (slice(10, 30, None), slice(10, 20, None)))
    assert bbox.get_overlap_slices((50, 50)) == slc


def test_bounding_box_extent():
    bbox = RegionBoundingBox(1, 10, 2, 20)
    assert_allclose(bbox.extent, (0.5, 9.5, 1.5, 19.5))


@pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
def test_bounding_box_as_artist():
    bbox = RegionBoundingBox(1, 10, 2, 20)
    patch = bbox.as_artist()

    assert_allclose(patch.get_xy(), (0.5, 1.5))
    assert_allclose(patch.get_width(), 9)
    assert_allclose(patch.get_height(), 18)


@pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
def test_bounding_box_plot():
    from matplotlib.patches import Patch
    bbox = RegionBoundingBox(1, 10, 2, 20)
    patch = bbox.plot()
    assert isinstance(patch, Patch)


@pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
def test_bounding_box_to_region():
    bbox = RegionBoundingBox(1, 10, 2, 20)
    region = RectanglePixelRegion(PixCoord(5.0, 10.5), width=9., height=18.)
    bbox_region = bbox.to_region()
    assert_allclose(bbox_region.center.xy, region.center.xy)
    assert bbox_region.width == region.width
    assert bbox_region.height == region.height
    assert bbox_region.angle == region.angle


def test_bounding_box_union():
    bbox1 = RegionBoundingBox(1, 10, 2, 20)
    bbox2 = RegionBoundingBox(5, 21, 7, 32)
    bbox_union_expected = RegionBoundingBox(1, 21, 2, 32)
    bbox_union1 = bbox1 | bbox2
    bbox_union2 = bbox1.union(bbox2)

    assert bbox_union1 == bbox_union_expected
    assert bbox_union1 == bbox_union2

    with pytest.raises(TypeError):
        bbox1.union((5, 21, 7, 32))


def test_bounding_box_intersect():
    bbox1 = RegionBoundingBox(1, 10, 2, 20)
    bbox2 = RegionBoundingBox(5, 21, 7, 32)
    bbox_intersect_expected = RegionBoundingBox(5, 10, 7, 20)
    bbox_intersect1 = bbox1 & bbox2
    bbox_intersect2 = bbox1.intersection(bbox2)

    assert bbox_intersect1 == bbox_intersect_expected
    assert bbox_intersect1 == bbox_intersect2

    with pytest.raises(TypeError):
        bbox1.intersection((5, 21, 7, 32))

    assert bbox1.intersection(RegionBoundingBox(30, 40, 50, 60)) is None



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module defines a class for region masks.
"""
import warnings

import astropy.units as u
import numpy as np
from astropy.utils import minversion

__all__ = ['RegionMask']

COPY_IF_NEEDED = False if not minversion(np, '2.0.0.dev') else None


class RegionMask:
    """
    Class for a region mask.

    Parameters
    ----------
    data : array_like
        A 2D array representing the fractional overlap of a region
        on the pixel grid. This should be the full-sized (i.e., not
        truncated) array that is the direct output of one of the
        low-level "geometry" functions.

    bbox : `regions.RegionBoundingBox`
        The bounding box object defining the region minimal bounding
        box.

    Examples
    --------
    Usage examples are provided in the :ref:`overlap-masks` section of
    the docs.
    """

    def __init__(self, data, bbox):
        self.data = np.asanyarray(data)
        if self.data.shape != bbox.shape:
            raise ValueError('mask data and bounding box must have the same '
                             'shape')
        self.bbox = bbox
        self._mask = (self.data == 0)

    def __array__(self, dtype=None, copy=COPY_IF_NEEDED):
        """
        Array representation of the mask data array (e.g., for
        matplotlib).
        """
        return np.array(self.data, dtype=dtype, copy=copy)

    @property
    def shape(self):
        """
        The shape of the mask data array.
        """
        return self.data.shape

    def get_overlap_slices(self, shape):
        """
        Get slices for the overlapping part of the region mask and a 2D
        array.

        Parameters
        ----------
        shape : 2-tuple of int
            The shape of the 2D array.

        Returns
        -------
        slices_large : tuple of slices or `None`
            A tuple of slice objects for each axis of the large array,
            such that ``large_array[slices_large]`` extracts the region
            of the large array that overlaps with the small array.
            `None` is returned if there is no overlap of the bounding
            box with the given image shape.

        slices_small : tuple of slices or `None`
            A tuple of slice objects for each axis of the region mask
            array such that ``small_array[slices_small]`` extracts the
            region that is inside the large array. `None` is returned if
            there is no overlap of the bounding box with the given image
            shape.
        """
        return self.bbox.get_overlap_slices(shape)

    def to_image(self, shape, dtype=float):
        """
        Return an image of the mask in a 2D array of the given shape,
        taking any edge effects into account.

        Parameters
        ----------
        shape : tuple of int
            The ``(ny, nx)`` shape of the output array.

        dtype : data-type, optional
            The desired data type for the array. This should be a
            floating data type if the `RegionMask` was created with the
            "exact" or "subpixel" mode, otherwise the fractional mask
            weights will be altered. A integer data type may be used if
            the `RegionMask` was created with the "center" mode.

        Returns
        -------
        result : `~numpy.ndarray`
            A 2D array of the mask.
        """
        if len(shape) != 2:
            raise ValueError('input shape must have 2 elements.')

        # find the overlap of the mask on the output image shape
        slices_large, slices_small = self.get_overlap_slices(shape)

        if slices_small is None:
            return None  # no overlap

        # insert the mask into the output image
        image = np.zeros(shape, dtype=dtype)
        image[slices_large] = self.data[slices_small]
        return image

    def cutout(self, data, fill_value=0.0, copy=False):
        """
        Create a cutout from the input data over the mask bounding box,
        taking any edge effects into account.

        Parameters
        ----------
        data : array_like
            A 2D array on which to apply the region mask.

        fill_value : float, optional
            The value used to fill pixels where the region mask does not
            overlap with the input ``data``.  The default is 0.

        copy : bool, optional
            If `True` then the returned cutout array will always be hold
            a copy of the input ``data``.  If `False` and the mask is
            fully within the input ``data``, then the returned cutout
            array will be a view into the input ``data``.  In cases
            where the mask partially overlaps or has no overlap with the
            input ``data``, the returned cutout array will always hold a
            copy of the input ``data`` (i.e., this keyword has no
            effect).

        Returns
        -------
        result : `~numpy.ndarray` or `None`
            A 2D array cut out from the input ``data`` representing the
            same cutout region as the region mask.  If there is a
            partial overlap of the region mask with the input data,
            pixels outside of the data will be assigned to
            ``fill_value``.  `None` is returned if there is no overlap
            of the region with the input ``data``.
        """
        data = np.asanyarray(data)
        if data.ndim != 2:
            raise ValueError('data must be a 2D array.')

        # find the overlap of the mask on the output image shape
        slices_large, slices_small = self.get_overlap_slices(data.shape)

        if slices_small is None:
            return None  # no overlap

        cutout_shape = (slices_small[0].stop - slices_small[0].start,
                        slices_small[1].stop - slices_small[1].start)

        if cutout_shape == self.shape:
            cutout = data[slices_large]
            if copy:
                cutout = np.copy(cutout)
            return cutout

        # cutout is always a copy for partial overlap
        dtype = float if ~np.isfinite(fill_value) else data.dtype
        cutout = np.zeros(self.shape, dtype=dtype)
        cutout[:] = fill_value
        cutout[slices_small] = data[slices_large]

        if isinstance(data, u.Quantity):
            cutout <<= data.unit

        return cutout

    def multiply(self, data, fill_value=0.0):
        """
        Multiply the region mask with the input data, taking any edge
        effects into account.

        The result is a mask-weighted cutout from the data.

        Parameters
        ----------
        data : array_like or `~astropy.units.Quantity`
            The 2D array to multiply with the region mask.

        fill_value : float, optional
            The value is used to fill pixels where the region mask does
            not overlap with the input ``data``.  The default is 0.

        Returns
        -------
        result : `~numpy.ndarray` or `None`
            A 2D mask-weighted cutout from the input ``data``. If
            there is a partial overlap of the region mask with the
            input data, pixels outside of the data will be assigned to
            ``fill_value`` before being multiplied with the mask. `None`
            is returned if there is no overlap of the region with the
            input ``data``.
        """
        cutout = self.cutout(data, fill_value=fill_value)
        if cutout is None:
            return None
        else:
            # ignore multiplication with non-finite data values
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', RuntimeWarning)
                weighted_cutout = cutout * self.data

            # fill values outside of the mask but within the bounding box
            weighted_cutout[self._mask] = fill_value

            return weighted_cutout

    def _get_overlap_cutouts(self, shape, mask=None):
        """
        Get the aperture mask weights, pixel mask, and slice for the
        overlap with the input shape.

        If input, the ``mask`` is included in the output pixel mask
        cutout.

        Parameters
        ----------
        shape : tuple of int
            The shape of data.

        mask : array_like (bool), optional
            A boolean mask with the same shape as ``shape`` where a
            `True` value indicates a masked pixel.

        Returns
        -------
        slices_large : tuple of slices or `None`
            A tuple of slice objects for each axis of the large array
            of given ``shape``, such that ``large_array[slices_large]``
            extracts the region of the large array that overlaps with
            the small array. `None` is returned if there is no overlap
            of the bounding box with the given image shape.

        aper_weights: 2D float `~numpy.ndarray`
            The cutout aperture mask weights for the overlap.

        pixel_mask: 2D bool `~numpy.ndarray`
            The cutout pixel mask for the overlap.

        Notes
        -----
        This method is separate from ``get_values`` to facilitate
        applying the same slices, aper_weights, and pixel_mask to
        multiple associated arrays (e.g., data and error arrays). It is
        used in this way by the `PixelAperture.do_photometry` method.
        """
        if mask is not None and mask.shape != shape:
            raise ValueError('mask and data must have the same shape')

        slc_large, slc_small = self.get_overlap_slices(shape)
        if slc_large is None:  # no overlap
            return None, None, None

        aper_weights = self.data[slc_small]
        pixel_mask = (aper_weights > 0)  # good pixels

        if mask is not None:
            pixel_mask &= ~mask[slc_large]

        return slc_large, aper_weights, pixel_mask

    def get_values(self, data, mask=None):
        """
        Get the mask-weighted pixel values from the data as a 1D array.

        If the `~regions.RegionMask` was created with ``mode='center'``,
        (where the mask weights are only 1 or 0), then the returned
        values will simply be pixel values extracted from the data.

        Parameters
        ----------
        data : array_like or `~astropy.units.Quantity`
            The 2D array from which to get mask-weighted values.

        mask : array_like (bool), optional
            A boolean mask with the same shape as ``data`` where a
            `True` value indicates the corresponding element of ``data``
            is not returned in the result.

        Returns
        -------
        result : `~numpy.ndarray`
            A 1D array of mask-weighted pixel values from the input
            ``data``. If there is no overlap of the region with the
            input ``data``, the result will be an empty array with shape
            (0,).
        """
        slc_large, aper_weights, pixel_mask = self._get_overlap_cutouts(
            data.shape, mask=mask)

        if slc_large is None:
            return np.array([])

        # ignore multiplication with non-finite data values
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            # pixel_mask is used so that pixels value where data = 0 and
            # aper_weights != 0 are still returned
            return (data[slc_large] * aper_weights)[pixel_mask]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the ds9 subpackage.
"""

import os
import warnings

import astropy.units as u
import pytest
from astropy.coordinates import Angle, SkyCoord
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.data import get_pkg_data_filenames
from astropy.utils.exceptions import AstropyUserWarning
from numpy.testing import assert_allclose, assert_equal

from regions._utils.optional_deps import HAS_MATPLOTLIB
from regions.core import PixCoord, Regions, RegionVisual
from regions.shapes import (CirclePixelRegion, CircleSkyRegion,
                            PointPixelRegion, RegularPolygonPixelRegion,
                            TextPixelRegion)
from regions.tests.helpers import assert_region_allclose


def test_roundtrip(tmpdir):
    filenames = get_pkg_data_filenames('data', pattern='*.reg')

    # AstropyUserWarning will be emitted only for some of the files
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', AstropyUserWarning)
        # Check that all test files are readable
        for filename in filenames:
            regions = Regions.read(filename, format='ds9')
            assert len(regions) > 0

            tempfile = tmpdir.join('tmp.ds9').strpath
            regions.write(tempfile, format='ds9', overwrite=True, precision=20)
            regions2 = Regions.read(tempfile, format='ds9')
            assert len(regions2) > 0
            for reg1, reg2 in zip(regions, regions2, strict=True):
                assert_region_allclose(reg1, reg2)


def test_serialize():
    """
    Simple test for serialize.
    """
    center = SkyCoord(42, 43, unit='deg')
    radius = Angle(3, 'deg')
    region = CircleSkyRegion(center, radius)
    expected = ('# Region file format: DS9 astropy/regions\nicrs\n'
                'circle(42.0000,43.0000,3.0000)\n')
    actual = region.serialize(format='ds9', precision=4)
    assert actual == expected


def test_serialize_parse_text():
    """
    Test serialization of Text region.
    """
    text = 'Example Text'
    region = TextPixelRegion(PixCoord(42, 43), text=text)
    expected = ('# Region file format: DS9 astropy/regions\nglobal '
                f'text={{{text}}}\nimage\ntext(43.0000,44.0000)\n')
    actual = region.serialize(format='ds9', precision=4)
    assert actual == expected

    reg = Regions.parse(actual, format='ds9')[0]
    assert reg.text == text
    assert 'text' not in reg.meta


def test_parse():
    """
    Simple test for parse.
    """
    region_str = ('# Region file format: DS9 astropy/regions\nfk5\n'
                  'circle(42.0000,43.0000,3.0000)\n')
    region = Regions.parse(region_str, format='ds9')[0]

    assert_allclose(region.center.ra.deg, 42)
    assert_allclose(region.center.dec.deg, 43)
    assert_quantity_allclose(region.radius, 3 * u.deg)


def test_exclude():
    """
    Simple parse test for an excluded region.
    """
    region_str = ('# Region file format: DS9 astropy/regions\nfk5\n'
                  ' -circle(42.0000,43.0000,3.0000)\n')
    region = Regions.parse(region_str, format='ds9')[0]
    assert not region.meta['include']


def test_read_write(tmpdir):
    """
    Simple test for write and read.
    """
    center = SkyCoord(42, 43, unit='deg', frame='fk5')
    radius = Angle(3, 'deg')
    region = CircleSkyRegion(center, radius)
    region.meta['text'] = 'ExampleText'

    filename = os.path.join(str(tmpdir), 'ds9.reg')
    region.write(filename, format='ds9')
    region2 = Regions.read(filename, format='ds9')[0]

    assert_quantity_allclose(region.center.ra, 42 * u.deg)
    assert_quantity_allclose(region.center.dec, 43 * u.deg)
    assert_quantity_allclose(region.radius, 3 * u.deg)
    assert 'text' in region2.meta
    assert region2.meta['text'] == 'ExampleText'


def test_invalid_region_warns():
    ds9_str = ('# Region file format: DS9 astropy/regions\nfk5\n'
               'circle(42.0000,43.0000,3.0000)\ninvalidregion(blah)')

    with pytest.warns(AstropyUserWarning) as warn_results:
        regions = Regions.parse(ds9_str, format='ds9')
    assert len(regions) == 1
    assert len(warn_results) == 1


def test_global_parser():
    """
    Test parsing global metadata.
    """
    region_str = ('global color=green dashlist=8 3 width=1 '
                  'font="helvetica 10 normal roman" select=1 highlite=1 '
                  'dash=0 fixed=0 edit=1 move=1 delete=1 include=1 source=1\n'
                  'icrs;circle(42.0000,43.0000,3.0000)')
    region = Regions.parse(region_str, format='ds9')[0]

    meta_expected = {'select': 1, 'highlite': 1, 'fixed': 0, 'edit': 1,
                     'move': 1, 'delete': 1, 'include': 1, 'source': 1}
    visual_expected = {'linewidth': 1, 'fontname': 'helvetica',
                       'fontsize': 10, 'fontweight': 'normal',
                       'fontstyle': 'normal', 'facecolor': 'green',
                       'edgecolor': 'green', 'default_style': 'ds9'}

    assert region.meta == meta_expected
    assert region.visual == visual_expected


def test_meta_color():
    region_str = ('# Region file format: DS9 astropy/regions\nicrs\n'
                  'circle(42.0000,43.0000,3.0000) # color=green\n'
                  'circle(43.0000,43.0000,3.0000) # color=orange\n')
    regions = Regions.parse(region_str, format='ds9')

    assert regions[0].visual['facecolor'] == 'green'
    assert regions[0].visual['edgecolor'] == 'green'
    assert regions[1].visual['facecolor'] == 'orange'
    assert regions[1].visual['edgecolor'] == 'orange'


def test_meta_color_override_global():
    """
    Color parsing test in the presence of a global color.
    """
    region_str = ('# Region file format: DS9 astropy/regions\n'
                  'global color=blue dashlist=8 3 width=1 '
                  'font="helvetica 10 normal roman" select=1 highlite=1 '
                  'dash=0 fixed=0 edit=1 move=1 delete=1 include=1 source=1\n'
                  'icrs\ncircle(42.0000,43.0000,3.0000) # color=green\n'
                  'circle(42.0000,43.0000,5.0000) # color=orange\n'
                  'circle(42.0000,43.0000,7.0000)')
    regions = Regions.parse(region_str, format='ds9')
    assert regions[0].visual['facecolor'] == 'green'
    assert regions[1].visual['facecolor'] == 'orange'
    assert regions[2].visual['facecolor'] == 'blue'


def test_issue134_regression():
    region_str = 'galactic; circle(+0:14:26.064,+0:00:45.206,30.400")'
    region = Regions.parse(region_str, format='ds9')[0]
    assert_quantity_allclose(region.radius, 30.4 * u.arcsec)


def test_issue65_regression():
    region_str = 'J2000; circle 188.5557102 12.0314056 1" # color=red'
    region = Regions.parse(region_str, format='ds9')[0]
    assert_quantity_allclose(region.center.ra, 188.5557102 * u.deg)
    assert_quantity_allclose(region.center.dec, 12.0314056 * u.deg)
    assert_quantity_allclose(region.radius, 1 * u.arcsec)


def test_frame_uppercase():
    """
    Regression test for issue #236 (PR #237).
    """
    region_str = ('GALACTIC\ncircle(188.5557102,12.0314056,7.1245) # '
                  'text={This message has both a " and : in it} '
                  'textangle=30')
    region = Regions.parse(region_str, format='ds9')[0]
    assert_quantity_allclose(region.center.l, 188.5557102 * u.deg)
    assert_quantity_allclose(region.center.b, 12.0314056 * u.deg)
    assert_quantity_allclose(region.radius, 7.1245 * u.deg)


def test_pixel_angle():
    """
    Check whether angle in PixelRegions is a u.Quantity object.
    """
    region_str = 'image\nbox(1.5,2,2,1,0)'
    region = Regions.parse(region_str, format='ds9')[0]
    assert isinstance(region.angle, u.Quantity)
    assert_quantity_allclose(region.angle, 0. * u.deg)


def test_parse_formats():
    region1_str = 'image\ncircle(1.5, 2, 2)'
    region1 = Regions.parse(region1_str, format='ds9')[0]
    region2_str = 'image\ncircle(1.5i, 2i, 2i)'
    region2 = Regions.parse(region2_str, format='ds9')[0]
    assert region1 == region2

    region3_str = 'icrs\ncircle(1.5, 2, 2)'
    region3 = Regions.parse(region3_str, format='ds9')[0]
    region4_str = 'icrs\ncircle(1.5d, 2d, 2d)'
    region4 = Regions.parse(region4_str, format='ds9')[0]
    assert region3 == region4

    region5_str = 'icrs\ncircle(1.5r, 1r, 2r)'
    region5 = Regions.parse(region5_str, format='ds9')[0]
    assert_quantity_allclose(region5.center.ra, 1.5 * u.radian)
    assert_quantity_allclose(region5.center.dec, 1.0 * u.radian)
    assert_quantity_allclose(region5.radius, 2.0 * u.radian)

    region6_str = 'icrs\ncircle(1:20:30, 2:3:7, 2)'
    region6 = Regions.parse(region6_str, format='ds9')[0]
    assert_quantity_allclose(region6.center.ra, 20.124999999999996 * u.deg)
    assert_quantity_allclose(region6.center.dec, 2.051944444444444 * u.deg)
    assert_quantity_allclose(region6.radius, 2. * u.deg)

    region7_str = 'icrs\ncircle(1h20m30s, 2d3m7s, 2d)'
    region7 = Regions.parse(region7_str, format='ds9')[0]
    assert_quantity_allclose(region7.center.ra, 20.124999999999996 * u.deg)
    assert_quantity_allclose(region7.center.dec, 2.051944444444444 * u.deg)
    assert_quantity_allclose(region7.radius, 2. * u.deg)


def test_angle_serialization():
    """
    Regression test for issue #223 to ensure Angle arcsec inputs are
    correctly converted to degrees.
    """
    region = CircleSkyRegion(SkyCoord(10, 20, unit='deg'), Angle(1, 'arcsec'))
    region_str = region.serialize(format='ds9', precision=6)

    expected = ('# Region file format: DS9 astropy/regions\nicrs\n'
                'circle(10.000000,20.000000,0.000278)\n')
    assert region_str == expected


def test_semicolon():
    """
    Regression test for issue #238 to allow semicolons in the text
    field.
    """
    region_str = ('galactic\n'
                  'circle(0.003, 0.1, 206.696") # text={S17; test} color=red\n'
                  'circle(0.1, -0.5, 360.148") # text={S19} color=red')

    regions = Regions.parse(region_str, format='ds9')
    assert len(regions) == 2

    regstr = ('galactic;'
              'circle(202.4,47.2,10.9) # text={test; test; A1} color=red;'
              'circle(103.8,40.6,3.6) # text="A, \'B\', ; {C}, D"'
              ' color=orange;'
              'circle(103.8,30.6,3.6) # text="A, \'B, ; C}, D" color=orange;'
              'circle(303.8,40.6,3.6) # text=\'A, "B", ; {C}, D\' color=cyan;'
              'circle(303.8,40.6,3.11);'
              'circle(303.8,30.6,3.6) # text=\'A, "B, ; {C, D\' color=cyan;'
              'circle(503.8,40.6,3.6) # text={A, "B", ; \'C\', D}'
              ' color=yellow\n'
              'circle(503.8,30.6,3.6) # text={A, "B, ; \'C, D} color=yellow;'
              '-ellipse(104.84,68.55,7.93",3.96",158.39) # dashlist=8 3'
              ' select=1 highlite=1 dash=0 fixed=0 edit=1 move=1 delete=1'
              ' source=1 text={Ellipse;Text} background color=#0ff width=1'
              ' font="helvetica 10 normal italic";'
              "ellipse(104.84,68.55,7.93',3.96',158.39) # dashlist=8 3"
              ' text="hello; world;";'
              'circle(1.5, 2, 2) # text={this_is_text};'
              'text(151.1,51.2) # text={text ; hi ; there;} textangle=45'
              ' color=orange font="helvetica 12 normal roman";'
              'text(151.1,51.2) # text  =  {text ; hi ; there;} textangle=45'
              ' color=orange font="helvetica 12 normal roman";'
              'circle(1.5, 2, 2) # text={text="hi;world";text;"tes\'t};'
              "circle(1.5, 2, 2) # text={text='hi;world';text;\"tes't};")

    regions = Regions.parse(regstr, format='ds9')
    assert len(regions) == 15


def test_parser_no_metadata():
    """
    Regression test for issue #259 to ensure regions without metadata
    are parsed correctly.
    """
    region1_str = 'galactic;circle(42,43,3)'
    region2_str = 'galactic;circle 42 43 3'
    region1 = Regions.parse(region1_str, format='ds9')[0]
    region2 = Regions.parse(region2_str, format='ds9')[0]

    assert isinstance(region1, CircleSkyRegion)
    assert isinstance(region2, CircleSkyRegion)
    assert_quantity_allclose(region1.radius, 3.0 * u.deg)
    assert_quantity_allclose(region2.radius, 3.0 * u.deg)


def test_spaces_metadata():
    """
    Regression test for spaces before or after "=" in region metadata.
    """
    regstr = ('galactic;'
              'circle(202.4,47.2,10.9) # text={a; test; test; A1} color=red;'
              'circle(202.4,47.2,10.9) # text ={a; test; test; A1} color= red;'
              'circle(202.4,47.2,10.9) # text= {a; test; test; A1} color =red;'
              'circle(202.4,47.2,10.9) # text = {a; test; test; A1}'
              ' color = red;')
    regions = Regions.parse(regstr, format='ds9')
    assert len(regions) == 4
    for i in range(1, len(regions)):
        assert regions[0].visual == regions[i].visual
        assert regions[0].meta == regions[i].meta

    regstr2 = ('galactic;'
               'text(151.1,51.2) # text={text ; hi ; there;} textangle= 45'
               ' color=orange font="helvetica 12 normal roman";'
               'text(151.1,51.2) # text  ={text ; hi ; there;} textangle =45'
               ' color=orange font="helvetica 12 normal roman";'
               'text(151.1,51.2) # text=  {text ; hi ; there;} textangle  =45'
               ' color   =   orange font="helvetica 12 normal roman";'
               'text(151.1,51.2) # text   =  {text ; hi ; there;} textangle=45'
               ' color   =orange font="helvetica 12 normal roman";')

    regions = Regions.parse(regstr2, format='ds9')
    assert len(regions) == 4
    for i in range(1, len(regions)):
        assert regions[0].visual == regions[i].visual
        assert regions[0].meta == regions[i].meta


@pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
def test_point_boxcircle():
    import matplotlib.path as mpath

    region_str = ('# Region file format: DS9 astropy/regions\n'
                  'image\n'
                  'point(101.000000,101.000000) # color=red\n'
                  'point(101.000000,201.000000) # point=boxcircle color=blue\n'
                  'point(101.000000,301.000000)')
    regions = Regions.parse(region_str, format='ds9')

    assert isinstance(regions[0].as_artist().get_marker(), mpath.Path)
    assert regions[0].as_artist().get_markeredgecolor() == 'red'
    assert isinstance(regions[1].visual['marker'], mpath.Path)
    assert regions[1].as_artist().get_markeredgecolor() == 'blue'
    assert isinstance(regions[2].as_artist().get_marker(), mpath.Path)
    assert regions[2].as_artist().get_markeredgecolor() == '#00ff00'


@pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
def test_compound_color():
    region_str = ('# Region file format: DS9 astropy/regions\n'
                  'image\n'
                  'annulus(651.0,301.0,60.0,90.0) # color=red')
    regions = Regions.parse(region_str, format='ds9')
    assert regions[0].as_artist().get_edgecolor() == (1., 0., 0., 1.)


@pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
def test_default_mpl_kwargs():
    region_str = ('# Region file format: DS9 astropy/regions\n'
                  'image\n'
                  'circle(101.0,101.0,3.0) # color=red\n'
                  'circle(101.0,301.0,3.0)\n'
                  'point(101.0,101.0) # color=red\n'
                  'point(101.0,301.0)\n'
                  'text(101.0,101.0) # text={Text} color=red\n'
                  'text(101.0,101.0) # text={Text}')
    regions = Regions.parse(region_str, format='ds9')

    # Patch
    assert regions[0].visual['default_style'] == 'ds9'
    assert regions[0].as_artist().get_edgecolor() == (1, 0, 0, 1)
    assert regions[1].as_artist().get_edgecolor() == (0, 1, 0, 1)
    for i in (0, 1):
        regions[i].visual['default_style'] = None
    assert regions[0].as_artist().get_edgecolor() == (1, 0, 0, 1)
    assert regions[1].as_artist().get_edgecolor() == (0, 0, 0, 1)

    # Line2D
    assert regions[2].visual['default_style'] == 'ds9'
    assert regions[2].as_artist().get_markeredgecolor() == 'red'
    assert regions[3].as_artist().get_markeredgecolor() == '#00ff00'
    for i in (2, 3):
        regions[i].visual['default_style'] = None
    assert regions[2].as_artist().get_markeredgecolor() == 'red'
    assert regions[3].as_artist().get_markeredgecolor() == 'C0'

    # Text
    assert regions[4].visual['default_style'] == 'ds9'
    assert regions[4].as_artist().get_color() == 'red'
    assert regions[5].as_artist().get_color() == '#00ff00'
    for i in (4, 5):
        assert regions[i].as_artist().get_va() == 'center'
        assert regions[i].as_artist().get_ha() == 'center'
    for i in (4, 5):
        regions[i].visual['default_style'] = None
        assert regions[i].as_artist().get_va() == 'baseline'
        assert regions[i].as_artist().get_ha() == 'left'
    assert regions[4].as_artist().get_color() == 'red'
    assert regions[5].as_artist().get_color() == 'black'


def test_annulus():
    region1_str = 'image\nannulus(1, 2, 3, 4)'
    regions1 = Regions.parse(region1_str, format='ds9')
    assert len(regions1) == 1
    assert_equal(regions1[0].center.xy, (0, 1))
    assert_equal(regions1[0].inner_radius, 3.0)
    assert_equal(regions1[0].outer_radius, 4.0)

    # multiple annuli
    region2_str = 'image\nannulus(1, 2, 3, 4, 5, 6)'
    regions2 = Regions.parse(region2_str, format='ds9')
    assert len(regions2) == 3
    assert_equal(regions2[0].center.xy, (0, 1))
    assert_equal(regions2[0].inner_radius, 3.0)
    assert_equal(regions2[0].outer_radius, 4.0)
    assert_equal(regions2[1].center.xy, (0, 1))
    assert_equal(regions2[1].inner_radius, 4.0)
    assert_equal(regions2[1].outer_radius, 5.0)
    assert_equal(regions2[2].center.xy, (0, 1))
    assert_equal(regions2[2].inner_radius, 5.0)
    assert_equal(regions2[2].outer_radius, 6.0)


def test_ellipse():
    region1_str = 'image\nellipse (1, 2, 3, 4, 5)'
    region1 = Regions.parse(region1_str, format='ds9')[0]
    assert_equal(region1.center.xy, (0, 1))
    assert_equal(region1.width, 6.0)
    assert_equal(region1.height, 8.0)
    assert_equal(region1.angle, 5.0 * u.deg)

    region2_str = 'icrs\nellipse (1, 2, 3, 4, 5)'
    region2 = Regions.parse(region2_str, format='ds9')[0]
    assert_equal(region2.center.ra, 1. * u.deg)
    assert_equal(region2.center.dec, 2. * u.deg)
    assert_equal(region2.width, 6.0 * u.deg)
    assert_equal(region2.height, 8.0 * u.deg)
    assert_equal(region2.angle, 5.0 * u.deg)

    # elliptical annulus
    region3_str = 'image\nellipse (1, 2, 3, 4, 5, 6, 7)'
    region3 = Regions.parse(region3_str, format='ds9')[0]
    assert_equal(region3.center.xy, (0, 1))
    assert_equal(region3.inner_width, 6.0)
    assert_equal(region3.inner_height, 8.0)
    assert_equal(region3.outer_width, 10.0)
    assert_equal(region3.outer_height, 12.0)
    assert_equal(region3.angle, 7.0 * u.deg)

    # multiple elliptical annuli
    region4_str = 'image\nellipse (1, 2, 3, 4, 5, 6, 7, 8, 9)'
    region4 = Regions.parse(region4_str, format='ds9')
    assert len(region4) == 2
    assert_equal(region4[0].center.xy, (0, 1))
    assert_equal(region4[0].inner_width, 6.0)
    assert_equal(region4[0].inner_height, 8.0)
    assert_equal(region4[0].outer_width, 10.0)
    assert_equal(region4[0].outer_height, 12.0)
    assert_equal(region4[0].angle, 9.0 * u.deg)
    assert_equal(region4[1].center.xy, (0, 1))
    assert_equal(region4[1].inner_width, 10.0)
    assert_equal(region4[1].inner_height, 12.0)
    assert_equal(region4[1].outer_width, 14.0)
    assert_equal(region4[1].outer_height, 16.0)
    assert_equal(region4[1].angle, 9.0 * u.deg)


def test_box():
    region1_str = 'image\nbox (1, 2, 3, 4, 5)'
    region1 = Regions.parse(region1_str, format='ds9')[0]
    assert_equal(region1.center.xy, (0, 1))
    assert_equal(region1.width, 3.0)
    assert_equal(region1.height, 4.0)
    assert_equal(region1.angle, 5.0 * u.deg)

    region2_str = 'icrs\nbox (1, 2, 3, 4, 5)'
    region2 = Regions.parse(region2_str, format='ds9')[0]
    assert_equal(region2.center.ra, 1. * u.deg)
    assert_equal(region2.center.dec, 2. * u.deg)
    assert_equal(region2.width, 3.0 * u.deg)
    assert_equal(region2.height, 4.0 * u.deg)
    assert_equal(region2.angle, 5.0 * u.deg)

    # box (rectangle) annulus
    region3_str = 'image\nbox (1, 2, 3, 4, 5, 6, 7)'
    region3 = Regions.parse(region3_str, format='ds9')[0]
    assert_equal(region3.center.xy, (0, 1))
    assert_equal(region3.inner_width, 3.0)
    assert_equal(region3.inner_height, 4.0)
    assert_equal(region3.outer_width, 5.0)
    assert_equal(region3.outer_height, 6.0)
    assert_equal(region3.angle, 7.0 * u.deg)

    # multiple box (rectangle) annuli
    region4_str = 'image\nbox (1, 2, 3, 4, 5, 6, 7, 8, 9)'
    region4 = Regions.parse(region4_str, format='ds9')
    assert len(region4) == 2
    assert_equal(region4[0].center.xy, (0, 1))
    assert_equal(region4[0].inner_width, 3.0)
    assert_equal(region4[0].inner_height, 4.0)
    assert_equal(region4[0].outer_width, 5.0)
    assert_equal(region4[0].outer_height, 6.0)
    assert_equal(region4[0].angle, 9.0 * u.deg)
    assert_equal(region4[1].center.xy, (0, 1))
    assert_equal(region4[1].inner_width, 5.0)
    assert_equal(region4[1].inner_height, 6.0)
    assert_equal(region4[1].outer_width, 7.0)
    assert_equal(region4[1].outer_height, 8.0)
    assert_equal(region4[1].angle, 9.0 * u.deg)


def test_invalid_metadata():
    # test that invalid ds9 metadata raises warnings
    regstr = ('# Region file format: DS9 version 4.1\n'
              'global color=green dashlist=8 3 width=1 '
              'font="helvetica 10 normal roman" select=1 highlite=1 dash=1 '
              'fixed=0 edit=1 move=1 rotate=1 delete=1 include=1 source=1.1 '
              'background=0 text={hello world} fill=1.8 point=diamond 11.2 '
              'textrotate=1 textangle=34 tag={Tag 1} tag={Tag 2} line=3 1\n'
              'image; circle(100.4,47.2,10.9) # color=red\n'
              'image; circle(202.4,47.2,10.9) # color=blue point=junk 10\n'
              'image; circle(302.4,147.2,10.9) # color=cyan point=invalid\n')

    with pytest.warns(AstropyUserWarning) as record:
        regs = Regions.parse(regstr, format='ds9')
        assert len(record) == 12
        assert 'source=1.1' in record[0].message.args[0]
        assert 'fill=1.8' in record[1].message.args[0]
        assert 'point=diamond 11.2' in record[2].message.args[0]
        assert 'line=3 1' in record[3].message.args[0]
        assert 'point=junk 10' in record[6].message.args[0]
        assert 'point=invalid' in record[10].message.args[0]

    assert len(regs) == 3


def test_unsupported_metadata():
    regstr = ('image; point(335.5,415.6) # point=diamond 11 color=yellow '
              'width=2 text={example} dash=1 tag={Tag 1} tag={Tag 2}')
    with pytest.warns(UserWarning) as record:
        Regions.parse(regstr, format='ds9')
        assert len(record) == 1
        assert 'dashed lines are unsupported' in record[0].message.args[0]


def test_text_metadata():
    # Regression test for issue #233: make sure that text metadata is
    # parsed and stored appropriately.
    region_str = 'image\ncircle(1.5, 2, 2) # text={this_is_text}'
    regions = Regions.parse(region_str, format='ds9')
    assert len(regions) == 1
    assert regions[0].meta['text'] == 'this_is_text'

    rstr = 'image; circle(503.6,490.6,31.1) # color=blue text={A, {B}, C}'
    reg = Regions.parse(rstr, format='ds9')[0]
    assert reg.meta['text'] == 'A, {B'

    rstr = 'image; circle(503.6,490.6,31.1) # color=blue text={A, {{B}}, C}'
    reg = Regions.parse(rstr, format='ds9')[0]
    assert reg.meta['text'] == 'A, {{B'

    rstr = "image; circle(503.6,490.6,31.1) # color=blue text='A, 'B', C'"
    reg = Regions.parse(rstr, format='ds9')[0]
    assert reg.meta['text'] == 'A, '

    rstr = 'image; circle(503.6,490.6,31.1) # color=blue text="A, "B", C"'
    reg = Regions.parse(rstr, format='ds9')[0]
    assert reg.meta['text'] == 'A, '


def test_mixed_coord():
    with pytest.raises(ValueError):
        CirclePixelRegion(PixCoord(10, 20), Angle(1, 'arcsec'))


def test_unsupported_marker():
    """
    Test that warning is issued when serializing a valid matplotlib
    marker, but unsupported by DS9.
    """
    region = PointPixelRegion(PixCoord(2, 2), visual=RegionVisual(marker='Z'))
    with pytest.warns(AstropyUserWarning):
        region.serialize(format='ds9')

    region = PointPixelRegion(PixCoord(2, 2),
                              visual=RegionVisual(marker='$f_{init}$'))
    with pytest.warns(AstropyUserWarning):
        region.serialize(format='ds9')


def test_serialize_regularpolygon():
    region = RegularPolygonPixelRegion(PixCoord(10, 10), 4, 20)
    poly_region = region.to_polygon()
    result1 = region.serialize(format='ds9')
    result2 = poly_region.serialize(format='ds9')
    assert result1 == result2


def test_serialize_empty_list():
    regions = Regions([])
    assert regions.serialize(format='ds9') == ''



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import warnings

from astropy.utils.exceptions import AstropyUserWarning

from regions.io.ds9.core import DS9ParserError, ds9_valid_symbols

__all__ = []


def _split_raw_metadata(raw_metadata):
    """
    Split the raw metadata into meta and visual dictionaries.

    The raw metadata is not changed, except unsupported metadata is
    ignored.

    Parameters
    ----------
    raw_metadata : dict
        The raw metadata as read from the DS9 region file/str.

    Returns
    -------
    meta, visual : dict
        The raw metadata and visual metadata dictionaries.
    """
    # not currently supported by regions
    unsupported_meta = ('line', 'vector', 'ruler', 'compass')

    # TODO: include text in visual keys to support text annotations for
    # all DS9 regions?
    ds9_visual_keys = ('color', 'dash', 'dashlist', 'fill', 'font', 'point',
                       'textangle', 'textrotate', 'width')

    meta = {}
    visual = {}
    for key, value in raw_metadata.items():
        if key in unsupported_meta:
            # don't warn for "line=0 0" (no arrows) because it works,
            # but skip adding it to metadata
            if key != 'line' or '1' in value:
                warnings.warn(f'DS9 meta "{key}={value}" is unsupported and '
                              'will be ignored', AstropyUserWarning)
            continue

        if key in ds9_visual_keys:
            visual[key] = value
        else:
            meta[key] = value

    # set the default plotting style to 'ds9' when parsing ds9 data
    visual['default_style'] = 'ds9'

    return meta, visual


def _translate_ds9_to_visual(shape, visual_meta):
    """
    Translate ds9 visual metadata to dictionary of matplotlib keywords.

    Parameters
    ----------
    shape : str
        The DS9 region shape (e.g., 'circle').

    visual_meta : dict
        The visual metadata.

    Returns
    -------
    visual : dict
        The visual dictionaries of region metadata translated
        to matplotlib keywords.
    """
    meta = visual_meta.copy()

    fill = meta.pop('fill', 0)
    # fill=1 is supported in DS9 only for the circle, ellipse, box, and
    # polygon shapes
    if fill == 1 and shape in ('circle', 'ellipse', 'box', 'polygon'):
        meta['fill'] = True

    dash = meta.pop('dash', 0)
    dashlist = meta.pop('dashlist', None)
    if int(dash) == 1:
        if dashlist is not None:
            dashes = tuple(int(i) for i in dashlist.split())
            meta['linestyle'] = (0, dashes)
        else:
            meta['linestyle'] = 'dashed'

        if shape == 'point':
            warnings.warn('dashed lines are unsupported for DS9 point '
                          'regions and will be ignored', AstropyUserWarning)
            meta.pop('linestyle', None)

    # "point=symbol [size]"; size is optional, symbol is not
    point = meta.pop('point', None)
    if point is not None:
        point_ = point.split()
        if len(point_) == 1:
            ds9_marker = point_[0]
        elif len(point_) == 2:
            ds9_marker, meta['markersize'] = point_
        else:
            raise ValueError(f'invalid point data "{point}"')
        meta['marker'] = ds9_valid_symbols[ds9_marker]

    font = meta.pop('font', None)
    if font is not None:
        vals = font.split()
        if len(vals) == 4:
            (meta['fontname'], meta['fontsize'], meta['fontweight'],
             meta['fontstyle']) = vals
        if len(vals) == 3:
            meta['fontname'], meta['fontsize'], meta['fontweight'] = vals
        if len(vals) == 2:
            meta['fontname'], meta['fontsize'] = vals
        if len(vals) == 1:
            meta['fontname'] = vals[0]

        # define default font values (helvetica 10 normal roman)
        if 'fontsize' not in meta:
            meta['fontsize'] = '10'
        if 'fontweight' not in meta:
            meta['fontweight'] = 'normal'
        if 'fontstyle' not in meta:
            meta['fontstyle'] = 'normal'

        # fontsize is a string
        try:
            meta['fontsize'] = int(meta['fontsize'])
        except ValueError:
            raise DS9ParserError('font size must be an integer, got '
                                 f'{meta["fontsize"]}') from None
        meta['fontstyle'] = meta['fontstyle'].replace('roman', 'normal')

    if shape == 'point':
        width = meta.pop('width', None)
        if width is not None:
            meta['markeredgewidth'] = width
    else:
        invalid = ('marker', 'markersize')
        for key in invalid:
            meta.pop(key, None)

    if shape == 'text':
        textangle = meta.pop('textangle', None)
        if textangle is not None:
            textrotate = meta.pop('textrotate', None)
            if textrotate != 0:  # rotate if None or 1
                meta['rotation'] = textangle

        # remove invalid mpl kwargs for matplotlib.text.Text
        invalid = ('linestyle', 'linewidth', 'fill')
        for key in invalid:
            meta.pop(key, None)
    else:
        # these kwarg are valid only for matplotlib.text.Text
        invalid = ('textangle', 'textrotate')
        for key in invalid:
            meta.pop(key, None)

    if shape not in ('point', 'line', 'text'):
        color = meta.pop('color', None)
        if color is not None:
            meta['facecolor'] = color
            meta['edgecolor'] = color

    return meta


def _remove_invalid_keys(region_meta, valid_keys):
    # TODO: instead of new dict, del region_meta in-place?
    meta = {}
    for key in region_meta:
        if key in valid_keys:
            meta[key] = region_meta[key]
    return meta


def _translate_metadata_to_ds9(region, shape):
    """
    Translate region metadata to valid ds9 meta keys.
    """
    meta = {**region.meta, **region.visual}
    # special case for Text regions
    if 'text' in region._params:
        meta = {'text': region.text, **meta}

    if 'annulus' in shape:
        # ds9 does not allow fill for annulus regions
        meta.pop('fill', None)

    fill = meta.pop('fill', None)
    if fill is not None:
        meta['fill'] = int(fill)

    if 'text' in meta:
        meta['text'] = f'{{{meta["text"]}}}'

    edgecolor = meta.pop('edgecolor', None)
    facecolor = meta.pop('facecolor', None)
    color = None
    if edgecolor is not None:
        color = edgecolor
    if facecolor is not None:
        if facecolor != color:
            warnings.warn('facecolor and edgecolor are different, edgecolor '
                          'will be used', AstropyUserWarning)
        if color is None:
            color = facecolor
    if color is not None:
        meta['color'] = color

    linewidth = meta.pop('linewidth', None)
    if linewidth is not None:
        meta['width'] = linewidth

    # point region marker width
    markeredgewidth = meta.pop('markeredgewidth', None)
    if markeredgewidth is not None:
        meta['width'] = markeredgewidth

    marker = meta.pop('marker', None)
    if marker is not None:
        symbol_map = {y: x for x, y in ds9_valid_symbols.items()}
        if marker in symbol_map:
            markersize = meta.pop('markersize', None)
            msize = ''
            if markersize is not None:
                msize = f' {markersize}'
            meta['point'] = f'{symbol_map[marker]}{msize}'
        else:
            warnings.warn(f'Unable to serialize marker "{marker}"',
                          AstropyUserWarning)

    fontname = meta.pop('fontname', None)
    if fontname is not None:
        fontsize = meta.pop('fontsize', 10)  # default 10
        fontweight = meta.pop('fontweight', 'normal')  # default normal
        # default roman
        fontstyle = meta.pop('fontstyle', 'roman').replace('normal', 'roman')
        meta['font'] = f'"{fontname} {fontsize} {fontweight} {fontstyle}"'

    linestyle = meta.pop('linestyle', None)
    if linestyle is not None:
        meta['dash'] = 1
    # if linestyle in ('dashed', '--'):
    if isinstance(linestyle, tuple):
        meta['dashlist'] = f'{linestyle[1][0]} {linestyle[1][1]}'

        # dashes = meta.pop('dashes', None)
        # if dashes is not None:
        #     meta['dashlist'] = f'{dashes[0]} {dashes[1]}'

    rotation = meta.pop('rotation', None)
    if rotation is not None:
        meta['textangle'] = rotation
        # meta['textrotate'] = 1

    # ds9 meta keys
    meta_keys = ['background', 'delete', 'edit', 'fixed', 'highlite',
                 'include', 'move', 'rotate', 'select', 'source', 'tag',
                 'text']
    visual_keys = ['color', 'dash', 'dashlist', 'fill', 'font', 'point',
                   'textangle', 'textrotate', 'width']
    valid_keys = meta_keys + visual_keys
    meta = _remove_invalid_keys(meta, valid_keys)

    return meta



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import abc
import copy
import operator

import numpy as np

from regions.core.metadata import RegionMeta, RegionVisual
from regions.core.pixcoord import PixCoord
from regions.core.registry import RegionsRegistry

__all__ = ['Region', 'PixelRegion', 'SkyRegion']
__doctest_skip__ = ['Region.serialize', 'Region.write']


class Region(abc.ABC):
    """
    Base class for all regions.
    """

    _params = ()

    def copy(self, **changes):
        """
        Make an independent (deep) copy.

        Parameters
        ----------
        **changes : dict
            Changes to make to the region parameters.
        """
        fields = list(self._params) + ['meta', 'visual']

        for field in fields:
            if field not in changes:
                changes[field] = copy.deepcopy(getattr(self, field))

        return self.__class__(**changes)

    def __repr__(self):
        prefix = f'{self.__class__.__name__}'
        cls_info = []
        if self._params is not None:
            for param in self._params:
                if param == 'text':
                    # place quotes around text value
                    keyval = f'{param}={getattr(self, param)!r}'
                else:
                    keyval = f'{param}={getattr(self, param)}'
                cls_info.append(keyval)
        cls_info = ', '.join(cls_info)
        return f'<{prefix}({cls_info})>'

    def __str__(self):
        cls_info = [('Region', self.__class__.__name__)]
        if self._params is not None:
            for param in self._params:
                if param == 'text':
                    # place quotes around text value
                    keyval = (param, repr(getattr(self, param)))
                else:
                    keyval = (param, getattr(self, param))
                cls_info.append(keyval)

        return '\n'.join([f'{key}: {val}' for key, val in cls_info])

    def __eq__(self, other):
        """
        Equality operator for Region.

        All Region properties are compared for strict equality except
        for Quantity parameters, which allow for different units if they
        are directly convertible.
        """
        if not isinstance(other, self.__class__):
            return False

        meta_params = ['meta', 'visual']
        self_params = list(self._params) + meta_params
        other_params = list(other._params) + meta_params

        # check that both have identical parameters
        if self_params != other_params:
            return False

        # now check the parameter values
        # Note that Quantity comparisons allow for different units
        # if they directly convertible (e.g., 1. * u.deg == 60. * u.arcmin)
        try:
            for param in self_params:
                # np.any is used for SkyCoord array comparisons
                if np.any(getattr(self, param) != getattr(other, param)):
                    return False
        except TypeError:
            # TypeError is raised from SkyCoord comparison when they do
            # not have equivalent frames. Here return False instead of
            # the TypeError.
            return False

        return True

    def __ne__(self, other):
        """
        Inequality operator for Region.

        Parameters
        ----------
        other : `Region`
            The other region that will be compared.
        """
        return not (self == other)

    @abc.abstractmethod
    def intersection(self, other):
        """
        Return a region representing the intersection of this region
        with ``other``.

        Parameters
        ----------
        other : `Region`
            The other region to use for the intersection.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def symmetric_difference(self, other):
        """
        Return the union of the two regions minus any areas contained in
        the intersection of the two regions.

        Parameters
        ----------
        other : `Region`
            The other region to use for the symmetric difference.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def union(self, other):
        """
        Return a region representing the union of this region with
        ``other``.

        Parameters
        ----------
        other : `Region`
            The other region to use for the union.
        """
        raise NotImplementedError

    def __and__(self, other):
        return self.intersection(other)

    def __or__(self, other):
        return self.union(other)

    def __xor__(self, other):
        return self.symmetric_difference(other)

    @classmethod
    def get_formats(cls):
        """
        Get the registered I/O formats as a Table.
        """
        return RegionsRegistry.get_formats(Region)

    def write(self, filename, format=None, overwrite=False, **kwargs):
        """
        Write the region to a region file in the specified format.

        This method allows writing a file in many supported data
        formats, e.g.,::

            >>> reg.write('new_regions.reg', format='ds9')
            >>> reg.write('new_regions.crtf', format='crtf')
            >>> reg.write('new_regions.fits', format='fits')

        A list of the available formats for `~regions.Region` is
        available using::

            >>> from regions import Region
            >>> Region.get_formats()

        Parameters
        ----------
        filename : str
            The filename or URL of the file to write.

        format : str, optional
            The file format specifier.

        overwrite : bool, optional
            If True, overwrite the output file if it exists. Raises an
            `OSError` if False and the output file exists. Default is
            False.

        **kwargs : dict, optional
            Keyword arguments passed to the data writer.
        """
        return RegionsRegistry.write([self], filename,
                                     Region, format=format,
                                     overwrite=overwrite, **kwargs)

    def serialize(self, format=None, **kwargs):
        """
        Serialize the region to a region string or table.

        This method allows serializing regions in many supported data
        formats, e.g.,::

            >>> reg1_str = reg.serialize(format='ds9')
            >>> reg2_str = reg.serialize(format='crtf')
            >>> reg3_tbl = reg.serialize(format='fits')

        A list of the available formats for `~regions.Region` is
        available using::

            >>> from regions import Region
            >>> Region.get_formats()

        Parameters
        ----------
        format : str, optional
            The file format specifier.

        **kwargs : dict, optional
            Keyword arguments passed to the data serializer.
        """
        return RegionsRegistry.serialize([self], Region, format=format,
                                         **kwargs)


class PixelRegion(Region):
    """
    Base class for all regions defined in pixel coordinates.
    """

    meta = RegionMeta()
    visual = RegionVisual()

    def intersection(self, other):
        """
        Return a region representing the intersection of this region
        with ``other``.

        Parameters
        ----------
        other : `Region`
            The other region to use for the intersection.
        """
        from regions.core.compound import CompoundPixelRegion
        return CompoundPixelRegion(region1=self, region2=other,
                                   operator=operator.and_)

    def symmetric_difference(self, other):
        """
        Return the union of the two regions minus any areas contained in
        the intersection of the two regions.

        Parameters
        ----------
        other : `Region`
            The other region to use for the symmetric difference.
        """
        from regions.core.compound import CompoundPixelRegion
        return CompoundPixelRegion(region1=self, region2=other,
                                   operator=operator.xor)

    def union(self, other):
        """
        Return a region representing the union of this region with
        ``other``.

        Parameters
        ----------
        other : `Region`
            The other region to use for the union.
        """
        from regions.core.compound import CompoundPixelRegion
        return CompoundPixelRegion(region1=self, region2=other,
                                   operator=operator.or_)

    @abc.abstractmethod
    def contains(self, pixcoord):
        """
        Check whether a position or positions fall inside the region.

        Parameters
        ----------
        pixcoord : `~regions.PixCoord`
            The position or positions to check.
        """
        raise NotImplementedError

    def __contains__(self, coord):
        if not coord.isscalar:
            raise ValueError(f'coord must be scalar. coord={coord}')
        return self.contains(coord)

    @abc.abstractmethod
    def to_sky(self, wcs):
        """
        Return a region defined in sky coordinates.

        Parameters
        ----------
        wcs : `~astropy.wcs.WCS`
            The world coordinate system transformation to use to convert
            from pixels to sky coordinates.

        Returns
        -------
        sky_region : `~regions.SkyRegion`
            The sky region.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def area(self):
        """
        The exact analytical area of the region shape.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def bounding_box(self):
        """
        The minimal bounding box (in integer pixel coordinates) that
        contains the region.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def to_mask(self, mode='center', subpixels=5):
        """
        Return a mask for the region.

        Parameters
        ----------
        mode : {'center', 'exact', 'subpixels'}, optional
            The method used to determine the overlap of the region on
            the pixel grid. Not all options are available for all region
            types. Note that the more precise methods are generally
            slower. The following methods are available:

                * ``'center'``:
                  A pixel is considered to be entirely in or out of the
                  region depending on whether its center is in or out of
                  the region. The returned mask will contain values only
                  of 0 (out) and 1 (in).

                * ``'exact'`` (default):
                  The exact fractional overlap of the region and each
                  pixel is calculated. The returned mask will contain
                  values between 0 and 1.

                * ``'subpixel'``:
                  A pixel is divided into subpixels (see the
                  ``subpixels`` keyword), each of which are considered
                  to be entirely in or out of the region depending
                  on whether its center is in or out of the region.
                  If ``subpixels=1``, this method is equivalent to
                  ``'center'``. The returned mask will contain values
                  between 0 and 1.

        subpixels : int, optional
            For the ``'subpixel'`` mode, resample pixels by this factor
            in each dimension. That is, each pixel is divided into
            ``subpixels ** 2`` subpixels.

        Returns
        -------
        mask : `~regions.RegionMask`
            A mask for the region.
        """
        raise NotImplementedError

    @staticmethod
    def _validate_mode(mode, subpixels):
        valid_modes = ('center', 'exact', 'subpixels')
        if mode not in valid_modes:
            raise ValueError(f'Invalid mask mode: {mode} (should be one '
                             f'of {valid_modes}')

        if (mode == 'subpixels'
                and (not isinstance(subpixels, int) or subpixels <= 0)):
            raise ValueError(f'Invalid subpixels value: {subpixels} '
                             '(should be a strictly positive integer)')

    @abc.abstractmethod
    def as_artist(self, origin=(0, 0), **kwargs):
        """
        Convert to matplotlib patch object for this region.

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` pixel position of the origin of the displayed
            image.

        **kwargs : dict
            Any keyword arguments accepted by
            `~matplotlib.patches.Patch`.

        Returns
        -------
        patch : `~matplotlib.patches.Patch`
            A matplotlib patch.
        """
        raise NotImplementedError

    def plot(self, origin=(0, 0), ax=None, **kwargs):
        """
        Plot the region on a matplotlib `~matplotlib.axes.Axes`
        instance.

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` pixel position of the origin of the displayed
            image.

        ax : `~matplotlib.axes.Axes` or `None`, optional
            The matplotlib axes on which to plot.  If `None`, then the
            current `~matplotlib.axes.Axes` instance is used.

        **kwargs : dict
            Any keyword arguments accepted by
            `~matplotlib.patches.Patch`.

        Returns
        -------
        artist : `matplotlib.artist.Artist`
            The matplotlib artist (typically a `~matplotlib.patches.Patch`
            object) for the plotted region. The artist can be used, for
            example, when adding a plot legend.
        """
        import matplotlib.pyplot as plt

        if ax is None:
            ax = plt.gca()

        artist = self.as_artist(origin=origin, **kwargs)
        ax.add_artist(artist)

        return artist


class SkyRegion(Region):
    """
    Base class for all regions defined in celestial coordinates.
    """

    def intersection(self, other):
        """
        Return a region representing the intersection of this region
        with ``other``.

        Parameters
        ----------
        other : `Region`
            The other region to use for the intersection.
        """
        from regions.core.compound import CompoundSkyRegion
        return CompoundSkyRegion(region1=self, region2=other,
                                 operator=operator.and_)

    def symmetric_difference(self, other):
        """
        Return the union of the two regions minus any areas contained in
        the intersection of the two regions.

        Parameters
        ----------
        other : `Region`
            The other region to use for the symmetric difference.
        """
        from regions.core.compound import CompoundSkyRegion
        return CompoundSkyRegion(region1=self, region2=other,
                                 operator=operator.xor)

    def union(self, other):
        """
        Return a region representing the union of this region with
        ``other``.

        Parameters
        ----------
        other : `Region`
            The other region to use for the union.
        """
        from regions.core.compound import CompoundSkyRegion
        return CompoundSkyRegion(region1=self, region2=other,
                                 operator=operator.or_)

    def contains(self, skycoord, wcs):
        """
        Check whether a sky coordinate falls inside the region.

        Parameters
        ----------
        skycoord : `~astropy.coordinates.SkyCoord`
            The position or positions to check.
        wcs : `~astropy.wcs.WCS`
            The world coordinate system transformation to use to convert
            between sky and pixel coordinates.
        """
        pixel_region = self.to_pixel(wcs)
        pixcoord = PixCoord.from_sky(skycoord, wcs)
        return pixel_region.contains(pixcoord)

    @abc.abstractmethod
    def to_pixel(self, wcs):
        """
        Return the equivalent region defined in pixel coordinates.

        Parameters
        ----------
        wcs : `~astropy.wcs.WCS`
            The world coordinate system transformation to use to convert
            between sky and pixel coordinates.

        Returns
        -------
        pixel_region : `~regions.PixelRegion`
            A pixel region.
        """
        raise NotImplementedError



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import itertools

from regions._utils.optional_deps import HAS_MATPLOTLIB
from regions.shapes import (CircleAnnulusPixelRegion, CircleAnnulusSkyRegion,
                            CirclePixelRegion, CircleSkyRegion,
                            EllipseAnnulusPixelRegion, EllipseAnnulusSkyRegion,
                            EllipsePixelRegion, EllipseSkyRegion,
                            LinePixelRegion, LineSkyRegion, PointPixelRegion,
                            PointSkyRegion, PolygonPixelRegion,
                            PolygonSkyRegion, RectangleAnnulusPixelRegion,
                            RectangleAnnulusSkyRegion, RectanglePixelRegion,
                            RectangleSkyRegion, TextPixelRegion, TextSkyRegion)

__all__ = []


def make_region_template():
    return itertools.chain(('coord', 'coord'), itertools.cycle(('length',)))


def make_polygon_template():
    return itertools.cycle(('coord',))


# mappings from DS9 frames to astropy coordinates frames
ds9_frame_map = {'image': 'image',
                 'icrs': 'icrs',
                 'fk5': 'fk5',
                 'j2000': 'fk5',
                 'fk4': 'fk4',
                 'b1950': 'fk4',
                 'galactic': 'galactic',
                 'ecliptic': 'barycentricmeanecliptic'}

# mappings from DS9 shape to region class
pixel_map = {'circle': CirclePixelRegion,
             'ellipse': EllipsePixelRegion,
             'box': RectanglePixelRegion,
             'polygon': PolygonPixelRegion,
             'annulus': CircleAnnulusPixelRegion,
             'ellipse_annulus': EllipseAnnulusPixelRegion,
             'rectangle_annulus': RectangleAnnulusPixelRegion,
             'line': LinePixelRegion,
             'point': PointPixelRegion,
             'text': TextPixelRegion}

sky_map = {'circle': CircleSkyRegion,
           'ellipse': EllipseSkyRegion,
           'box': RectangleSkyRegion,
           'polygon': PolygonSkyRegion,
           'annulus': CircleAnnulusSkyRegion,
           'ellipse_annulus': EllipseAnnulusSkyRegion,
           'rectangle_annulus': RectangleAnnulusSkyRegion,
           'line': LineSkyRegion,
           'point': PointSkyRegion,
           'text': TextSkyRegion}

ds9_shape_to_region = {}
ds9_shape_to_region['pixel'] = pixel_map
ds9_shape_to_region['sky'] = sky_map


ds9_params_template = {'point': ('coord', 'coord'),
                       'text': ('coord', 'coord'),
                       'circle': ('coord', 'coord', 'length'),
                       'line': ('coord', 'coord', 'coord', 'coord'),
                       'ellipse': make_region_template(),
                       'annulus': make_region_template(),
                       'box': make_region_template(),
                       'polygon': make_polygon_template()}


# mapping from regions shapes to ds9 shape formats
# unsupported ds9 shapes:
# vector, ruler, compass, projection, panda, epanda, bpanda
ds9_shape_templates = {'circle': ('circle',
                                  '{center},{radius}'),
                       'ellipse': ('ellipse',
                                   '{center},{width},{height}'
                                   ',{angle}'),
                       'rectangle': ('box',
                                     '{center},{width},{height},{angle}'),
                       'circleannulus': ('annulus',
                                         '{center},{inner_radius},'
                                         '{outer_radius}'),
                       'ellipseannulus': ('ellipse',
                                          '{center},{inner_width},'
                                          '{inner_height},'
                                          '{outer_width},'
                                          '{outer_height},{angle}'),
                       'rectangleannulus': ('box',
                                            '{center},{inner_width},'
                                            '{inner_height},{outer_width},'
                                            '{outer_height},{angle}'),
                       'polygon': ('polygon',
                                   '{vertices}'),
                       'line': ('line',
                                '{start},{end}'),
                       'point': ('point', '{center}'),
                       'text': ('text', '{center}')}


if not HAS_MATPLOTLIB:
    boxcircle = '8'  # octagon
    arrow = '^'  # triangle
else:
    import matplotlib.path as mpath

    vertices = [[0., -1.], [0.2652031, -1.],
                [0.51957987, -0.89463369], [0.70710678, -0.70710678],
                [0.89463369, -0.51957987], [1., -0.2652031],
                [1., 0.], [1., 0.2652031], [0.89463369, 0.51957987],
                [0.70710678, 0.70710678], [0.51957987, 0.89463369],
                [0.2652031, 1.], [0., 1.], [-0.2652031, 1.],
                [-0.51957987, 0.89463369], [-0.70710678, 0.70710678],
                [-0.89463369, 0.51957987], [-1., 0.2652031],
                [-1., 0.], [-1., -0.2652031],
                [-0.89463369, -0.51957987],
                [-0.70710678, -0.70710678],
                [-0.51957987, -0.89463369], [-0.2652031, -1.],
                [0., -1.], [0., -1.], [0., -1.], [1., -1.], [1., 1.],
                [-1., 1.], [-1., -1.], [0., -1.]]
    codes = [1, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
             4, 4, 4, 4, 4, 4, 4, 79, 1, 2, 2, 2, 2, 79]
    boxcircle = mpath.Path(vertices, codes)

    arrow_verts = [[-1, 0], [0, 0], [-1, 1], [0, 0], [0, 1]]
    arrow = mpath.Path(arrow_verts, codes=None)


# mapping from ds9 point symbols to matplotlib marker symbols
ds9_valid_symbols = {'circle': 'o',
                     'box': 's',
                     'diamond': 'D',
                     'x': 'x',
                     'cross': '+',
                     'arrow': arrow,
                     'boxcircle': boxcircle}


class DS9ParserError(Exception):
    """
    A custom exception for DS9 parsing errors.
    """



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import warnings
from dataclasses import dataclass

import astropy.units as u
import numpy as np
from astropy.io import fits
from astropy.table import QTable
from astropy.utils.exceptions import AstropyUserWarning

from regions.core import Region, Regions, SkyRegion
from regions.core.registry import RegionsRegistry
from regions.shapes import RegularPolygonPixelRegion

__all__ = []


@RegionsRegistry.register(Region, 'serialize', 'fits')
@RegionsRegistry.register(Regions, 'serialize', 'fits')
def _serialize_fits(regions):
    region_data = []
    for region in regions:
        if isinstance(region, SkyRegion):
            warnings.warn('Sky regions cannot be serialized to the FITS '
                          'region format, skipping.', AstropyUserWarning)
            continue
        regdata = _serialize_region_fits(region)
        if regdata is not None:
            region_data.append(regdata)

    if region_data:
        fits_table = _make_table(region_data)
        return fits_table
    return QTable()


@RegionsRegistry.register(Region, 'write', 'fits')
@RegionsRegistry.register(Regions, 'write', 'fits')
def _write_fits(regions, filename, header=None, overwrite=False):
    """
    Convert a list of `~regions.Region` to a FITS region table and write
    to a file.

    Parameters
    ----------
    regions : list
        A list of `~regions.Region` objects.

    filename : str
        The filename in which the table is to be written.

    header : `~astropy.io.fits.Header`, optional
        The FITS header.

    overwrite : bool, optional
        If True, overwrite the output file if it exists. Raises an
        `OSError` if False and the output file exists. Default is False.
    """
    output = _serialize_fits(regions)

    if header is None:
        hdudoc = ('ASC-FITS-REGION-1.2: Rots, McDowell: FITS REGION '
                  'Binary Table Design')
        header = dict([('EXTNAME', 'REGION'),
                       ('EXTVER', 1),
                       ('EXTLEVEL', 1),
                       ('HDUNAME', 'REGION'),
                       ('HDUCLASS', 'ASC'),
                       ('HDUCLAS1', 'REGION'),
                       ('HDUCLAS2', 'STANDARD'),
                       ('HDUVERS', '1.2.0'),
                       ('HDUDOC', hdudoc),
                       ('CONTENT', 'REGION'),
                       ('ORIGIN', 'astropy/regions')])

    bin_table = fits.BinTableHDU(data=output, header=header)
    bin_table.writeto(filename, overwrite=overwrite)


@dataclass
class _RegionData:
    """
    Class to hold region data.
    """

    shape: str
    x: np.ndarray
    y: np.ndarray
    r: np.ndarray
    rotang: np.ndarray
    component: int


def _serialize_region_fits(region):
    if isinstance(region, RegularPolygonPixelRegion):
        region = region.to_polygon()

    region_clsname = region.__class__.__name__

    unsupported_regions = ('RectangleAnnulusPixelRegion',
                           'LinePixelRegion', 'TextPixelRegion',
                           'CompoundPixelRegion')
    if region_clsname in unsupported_regions:
        warnings.warn(f'({region_clsname} cannot be serialized using the '
                      'FITS format, skipping.', AstropyUserWarning)
        return None

    # translate region class to FITS shape name
    shape = region_clsname.lower().replace('pixelregion', '')

    if region.meta.get('include', None) == 0:
        shape = f'!{shape}'

    region_map = {'circleannulus': 'annulus',
                  'ellipseannulus': 'elliptannulus',
                  'rectangle': 'rotbox'}
    if shape in region_map:
        shape = region_map[shape]

    shape_params = []
    rotang = None
    for param in region._params:
        value = getattr(region, param)
        if param in ('center', 'vertices'):
            x, y = value.xy
        elif param == 'angle':
            rotang = value
        else:
            # ellipse region is defined by full axis lengths, but
            # FITS regions file uses semi-axis lengths
            if shape == 'ellipse':
                value /= 2.0
            shape_params.append(value)

    if not shape_params:
        shape_params = 0
    if rotang is None:
        rotang = u.Quantity(0, 'deg')

    component = region.meta.get('component', None)

    return _RegionData(shape, np.atleast_1d(x), np.atleast_1d(y),
                       np.atleast_1d(shape_params), np.atleast_1d(rotang),
                       component)


def _make_column(arrays):
    arr_sizes = [arr.size for arr in arrays]
    arr_size = np.max(arr_sizes)

    data = []
    for (arr, size) in zip(arrays, arr_sizes, strict=True):
        pad_width = arr_size - size
        if pad_width != 0:
            arr = np.pad(arr, (0, pad_width), mode='constant')
        if not isinstance(arr[0], u.Quantity):
            arr <<= u.pix
        if arr_size == 1 and pad_width == 0:
            arr = arr[0]  # use scalars instead of len-1 arrays
        data.append(arr)

    data = u.Quantity(data)

    return data


def _define_components(region_data):
    components = np.array([regdata.component for regdata in region_data])

    none_idx = [i for i, item in enumerate(components) if item is None]
    if not none_idx:
        return components

    comps = [i for i in components if i is not None]
    if comps:
        start_component = np.max(comps) + 1
        components[none_idx] = np.arange(len(none_idx)) + start_component
    else:
        # all components are set to None - do not write a COMPONENT
        # column
        components = None

    return components


def _make_table(region_data):
    tbl = QTable()
    attrs = ('shape', 'x', 'y', 'r', 'rotang')

    for attr in attrs:
        arrays = [getattr(data, attr) for data in region_data]
        if attr not in ('shape',):
            arrays = _make_column(arrays)
        tbl[attr.upper()] = arrays

    components = _define_components(region_data)
    if components is not None:
        tbl['COMPONENT'] = components

    return tbl



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module proves classes to handle region metadata.
"""
from copy import deepcopy

__all__ = ['Meta', 'RegionMeta', 'RegionVisual']


class Meta(dict):
    """
    A base class for region metadata.

    Parameters
    ----------
    seq : dict-like, optional
        A dictionary or other mapping object to initialize the metadata.

    **kwargs
        Additional keyword arguments to initialize the metadata.
    """

    valid_keys = []
    key_mapping = {}

    def __init__(self, seq=None, **kwargs):
        super().__init__()

        if seq:
            if isinstance(seq, dict):
                for key, val in seq.items():
                    self.__setitem__(key, val)
            else:
                for key, val in seq:
                    self.__setitem__(key, val)

        if len(kwargs) > 0:
            for key, val in kwargs.items():
                self.__setitem__(key, val)

    def __setitem__(self, key, value):
        key = self.key_mapping.get(key, key)
        if key in self.valid_keys:
            super().__setitem__(key, value)
        else:
            raise KeyError(f'{key} is not a valid key for this class.')

    def __getitem__(self, item):
        item = self.key_mapping.get(item, item)
        return super().__getitem__(item)

    def update(self, *args, **kwargs):
        """
        Update the metadata with the input dictionary.

        Parameters
        ----------
        *args : dict-like
            A dictionary or other mapping object to update the metadata.

        **kwargs : dict-like
            Additional keyword arguments to update the metadata.
        """
        if args:
            if len(args) > 1:
                raise ValueError('Only one argument can be input')
            other = dict(args[0])
            for key, value in other.items():
                self[key] = value
        for key, value in kwargs.items():
            self[key] = value

    def setdefault(self, key, value=None):
        if key not in self:
            self[key] = value
        return self[key]

    def copy(self):
        """
        Make a deep copy of this object.
        """
        return deepcopy(self)


class RegionMeta(Meta):
    """
    A dictionary subclass that holds the meta attributes of the region.
    """

    valid_keys = ['background', 'comment', 'component', 'composite', 'corr',
                  'delete', 'edit', 'fixed', 'frame', 'highlite', 'include',
                  'label', 'line', 'move', 'name', 'range', 'restfreq',
                  'rotate', 'select', 'source', 'tag', 'text', 'textrotate',
                  'type', 'veltype']

    key_mapping = {}


class RegionVisual(Meta):
    """
    A dictionary subclass which holds the visual attributes of the
    region.
    """

    valid_keys = ['color', 'dash', 'dashlist', 'fill', 'font',
                  'fontname', 'fontsize', 'fontstyle', 'fontweight',
                  'labeloff', 'labelpos', 'labelcolor', 'line', 'linestyle',
                  'linewidth', 'marker', 'markersize', 'symbol', 'symsize',
                  'symthick', 'textangle', 'textrotate', 'usetex',
                  'default_style', 'dashes', 'markeredgewidth', 'rotation',
                  'facecolor', 'edgecolor']

    key_mapping = {'point': 'symbol', 'width': 'linewidth'}

    def _define_default_mpl_kwargs(self, artist):
        """
        Define the default matplotlib kwargs for the specified artist.

        The kwargs depend on the value of self.visual['default_style'],
        which can be set when reading region files. If this keywords is
        not set or set to 'mpl' or `None`, then the matplotlib defaults
        will be used, with the exception fill is turned off for Patch
        and Line2D objects.

        Parameters
        ----------
        artist : {'Text', 'Line2D', 'Patch'}
            The matplotlib artist type.

        Returns
        -------
        result : dict
            A dictionary of matplotlib keyword arguments.
        """
        kwargs = {}

        default_style = self.get('default_style', None)
        if default_style is None or default_style == 'mpl':
            # do not fill by default, which is the only change from
            # matplotlib defaults
            if artist == 'Patch':
                kwargs['fill'] = False
            elif artist == 'Line2D':
                kwargs['fillstyle'] = 'none'
                kwargs['marker'] = 'o'
            return kwargs

        # 'ds9' style is set when reading from ds9 region files
        elif default_style == 'ds9':
            kwargs['color'] = '#00ff00'  # green
            if artist == 'Text':
                kwargs['ha'] = 'center'  # text horizontal alignment
                kwargs['va'] = 'center'  # text vertical alignment
            elif artist == 'Line2D':
                from regions.io.ds9.core import ds9_valid_symbols

                kwargs['marker'] = ds9_valid_symbols['boxcircle']
                kwargs['markersize'] = 11
                kwargs['markeredgecolor'] = kwargs.pop('color')
                kwargs['fillstyle'] = 'none'
            elif artist == 'Patch':
                kwargs['edgecolor'] = kwargs.pop('color')
                kwargs['fill'] = False

            else:
                raise ValueError('invalid visual["default"] value')

        return kwargs

    def _to_mpl_kwargs(self, artist):
        """
        Convert the visual metadata to a dictionary of matplotlib
        keyword arguments for the given artist.

        Parameters
        ----------
        artist : {'Text', 'Line2D', 'Patch'}
            The matplotlib artist type.

        Returns
        -------
        result : dict
            A dictionary of matplotlib keyword arguments.
        """
        if artist == 'Text':
            keymap = {'font': 'family',
                      'fontstyle': 'style',
                      'fontweight': 'weight',
                      'fontsize': 'size',
                      'textangle': 'rotation'}

        elif artist == 'Line2D':
            keymap = {'symsize': 'markersize',
                      'color': 'markeredgecolor',
                      'linewidth': 'markeredgewidth',
                      'fill': 'fillstyle'}

        elif artist == 'Patch':
            keymap = {'color': 'edgecolor',
                      'fill': 'fill'}

        else:
            raise ValueError('invalid artist type')

        kwargs = {}
        for name, val in self.items():
            if name in keymap:
                # NOTE: this will override existing mpl kwargs
                kwargs[keymap[name]] = val
            else:
                kwargs[name] = val

        default_style = kwargs.pop('default_style', None)
        if default_style == 'ds9':
            for key, val in kwargs.items():
                if val == 'green':
                    # X11/mpl green is #008000, ds9 uses #00ff00
                    kwargs[key] = '#00ff00'

        return kwargs

    def define_mpl_kwargs(self, artist):
        """
        Define a dictionary of matplotlib keywords for the input
        ``artist`` from the region's ``visual`` properties.

        Parameters
        ----------
        artist : {'Text', 'Line2D', 'Patch'}
            The matplotlib artist type.

        Returns
        -------
        result : dict
            A dictionary of matplotlib keyword arguments.
        """
        if artist not in ('Patch', 'Line2D', 'Text'):
            raise ValueError(f'artist "{artist}" is not supported')

        kwargs = self._define_default_mpl_kwargs(artist)
        kwargs.update(self._to_mpl_kwargs(artist))

        remove_keys = []
        if artist != 'Text':
            remove_keys.extend(['fontname', 'fontsize', 'fontweight',
                                'fontstyle'])
        else:
            remove_keys.extend(['linewidth'])

        for key in remove_keys:
            kwargs.pop(key, None)

        if 'fontstyle' in kwargs:
            kwargs['fontstyle'].replace('roman', 'normal')

        return kwargs



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import warnings
from copy import deepcopy

from astropy.coordinates import Angle, SkyCoord
from astropy.units import Quantity
from astropy.utils.exceptions import AstropyUserWarning

from regions.core import (CompoundPixelRegion, CompoundSkyRegion, PixCoord,
                          PixelRegion, Region, Regions)
from regions.core.registry import RegionsRegistry
from regions.io.ds9.core import ds9_frame_map, ds9_shape_templates
from regions.io.ds9.meta import _translate_metadata_to_ds9
from regions.shapes import RegularPolygonPixelRegion

__all__ = []


@RegionsRegistry.register(Region, 'serialize', 'ds9')
@RegionsRegistry.register(Regions, 'serialize', 'ds9')
def _serialize_ds9(regions, precision=8):
    if not regions:
        return ''

    region_data = []
    for region in regions:
        if isinstance(region, (CompoundPixelRegion, CompoundSkyRegion)):
            warnings.warn('Cannot serialize a compound region, skipping',
                          AstropyUserWarning)

        if isinstance(region, RegularPolygonPixelRegion):
            region = region.to_polygon()

        region_data.append(_serialize_region_ds9(region, precision=precision))

    # ds9 file header
    output = '# Region file format: DS9 astropy/regions\n'

    # extract common region metadata and place in the global metadata
    all_meta = []
    for region in region_data:
        region_meta = deepcopy(region['meta'])
        region_meta.pop('tag', None)  # "tag" cannot be in global metadata
        all_meta.append(region_meta)

    global_meta = dict(set.intersection(*[set(meta_dict.items())
                                          for meta_dict in all_meta]))
    if global_meta:
        output += f'global {_make_meta_str(global_meta)}\n'

    # define region frame and metadata (removing items that are
    # defined in global metadata)
    frames = []
    metadata = []
    for region in region_data:
        frames.append(region['frame'])
        meta_tmp = region['meta']
        for key in global_meta:
            meta_tmp.pop(key)
        metadata.append(meta_tmp)

    # extract the coordinate frame if it is identical for all regions
    # TODO: extract common coord frame(s) for block(s) of consecutive regions
    frames = set(frames)
    global_frame = None
    if len(frames) == 1:
        global_frame = frames.pop()
        output += f'{global_frame}\n'

    # add line for each region
    for region, region_meta in zip(region_data, metadata, strict=True):
        if global_frame is None:
            output += f'{region["frame"]}; '

        meta_str = _make_meta_str(region_meta)
        if meta_str:
            meta_str = f' # {meta_str}'

        output += f'{region["region"]}{meta_str}\n'

    return output


@RegionsRegistry.register(Region, 'write', 'ds9')
@RegionsRegistry.register(Regions, 'write', 'ds9')
def _write_ds9(regions, filename, *, precision=8, overwrite=False):
    """
    Convert a list of `~regions.Region` to a DS9 string and write to a
    file.

    Parameters
    ----------
    regions : list
        A list of `~regions.Region` objects.

    filename : str
        The output region filename.

    precision : int, optional
        The level of decimal precision given as the number of decimal
        places.

    overwrite : bool, optional
        If `True`, overwrite the output file if it exists. If `False`
        (default) and the output file exists, an `OSError` is raised.
    """
    if os.path.lexists(filename) and not overwrite:
        raise OSError(f'{filename} already exists')

    output = _serialize_ds9(regions, precision=precision)
    with open(filename, 'w') as fh:
        fh.write(output)


def _get_region_shape(region):
    shape = region.__class__.__name__.lower().replace('skyregion', '')
    shape = shape.replace('pixelregion', '')
    return shape


def _get_frame_name(region, mapping):
    if isinstance(region, PixelRegion):
        frame = 'image'
    elif 'center' in region._params:
        frame = region.center.frame.name
    elif 'vertices' in region._params:
        frame = region.vertices.frame.name
    elif 'start' in region._params:
        frame = region.start.frame.name
    else:
        raise ValueError(f'Unable to get coordinate frame for {region!r}')

    if frame not in mapping:
        warnings.warn(f'Cannot serialize region with frame={frame}, skipping',
                      AstropyUserWarning)

    return mapping[frame]


def _make_meta_str(meta):
    metalist = []
    for key, val in meta.items():
        if key == 'tag':  # can have multiple tags; value is always a list
            metalist.append(' '.join([f'tag={{{val}}}' for val in val]))
        else:
            metalist.append(f'{key}={val}')
    return ' '.join(metalist)


def _get_region_params(region, shape_template, precision=8):
    ellipse_axes = ('width', 'height', 'inner_width', 'inner_height',
                    'outer_width', 'outer_height')
    ellipse_names = ('ellipse', 'ellipseannulus')

    param = {}
    for param_name in region._params:
        if param_name in ('text',):
            continue

        value = getattr(region, param_name)

        # DS9 ellipse parameters are serialized as semi-axis lengths, but
        # ellipse region is defined by full axis lengths
        is_ellipse = (shape_template[0] in ellipse_names
                      and param_name in ellipse_axes)
        if not isinstance(value, (PixCoord, SkyCoord)) and is_ellipse:
            # deepcopy to prevent changing value in memory
            value = deepcopy(value) / 2.0  # semi-axis lengths

        if isinstance(value, PixCoord):
            # pixels; ds9's origin is (1, 1)
            if value.isscalar:
                value = (f'{value.x + 1:0.{precision}f},'
                         f'{value.y + 1:0.{precision}f}')
            else:
                value_str = ''
                for val in value:
                    value_str += (f'{val.x + 1:0.{precision}f},'
                                  f'{val.y + 1:0.{precision}f},')
                value = value_str[:-1]

        elif isinstance(value, SkyCoord):
            val = value.to_string(precision=precision)
            # polygon region has multiple SkyCoord
            value = ' '.join(val) if not value.isscalar else val
            value = value.replace(' ', ',')

        elif isinstance(value, Angle):
            value = value.to_string(unit='deg', decimal=True,
                                    precision=precision)

        elif isinstance(value, Quantity):
            # [:-4] to trim ' deg' from string end
            value = value.to_string(unit='deg', precision=precision)[:-4]

        else:
            value = f'{value:.{precision}f}'

        param[param_name] = value

    try:
        param_str = shape_template[1].format(**param)
    except KeyError as err:
        raise ValueError(
            f'Unable to get shape parameters for {region!r}') from err

    return param_str


def _serialize_region_ds9(region, precision=8):
    frame_mapping = {v: k for k, v in ds9_frame_map.items()}
    frame = _get_frame_name(region, mapping=frame_mapping)

    shape = _get_region_shape(region)
    if shape not in ds9_shape_templates:
        warnings.warn(f'Cannot serialize region shape "{shape}", '
                      'skipping', AstropyUserWarning)

    region_params = _get_region_params(region,
                                       ds9_shape_templates[shape],
                                       precision=precision)

    region_type = ds9_shape_templates[shape][0]
    region_str = f'{region_type}({region_params})'

    region_meta = _translate_metadata_to_ds9(region, shape)

    return {'frame': frame, 'region': region_str, 'meta': region_meta}
